# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import enum
import textwrap
import typing

import dacite
import jsonpath_ng

from ci.util import not_none
from model import NamedModelElement
import concourse.model.traits.component_descriptor
import ocm
import oci.model as om

from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PrivilegeMode,
    PullRequestNotificationPolicy,
)
from concourse.model.base import (
  AttributeSpec,
  AttribSpecMixin,
  ScriptType,
  Trait,
  TraitTransformer,
)
from model.base import (
  ModelDefaultsMixin,
  ModelValidationError,
)


@dataclasses.dataclass
class TargetSpec:
    '''
    OCI Image build target:
    - target: the docker build target
    - image: the image push target (to where build result should be published)
    - name: image name (for use in component-descriptor)

    if name is not passed, a fallback is calculated from <publish.dockerimages.<name>> + target.
    '''
    target: str
    image: str
    name: str | None = None # default value is injected after deserialisation


IMG_DESCRIPTOR_ATTRIBS = (
    AttributeSpec.optional(
        name='registry',
        default=None,
        type=str,
        doc='name of the registry config to use when pushing the image.',
    ),
    AttributeSpec.optional(
        name='image',
        type=str,
        doc='''\
        image reference to publish the created container image to. required if on `targets` are set
        ''',
        default=None,
    ),
    AttributeSpec.optional(
        name='extra_push_targets',
        default=[],
        type=list[str],
        doc='''
        additional targets to publish built images to. Entries _may_ contain a tag (which is
        honoured, if present). Entries without a tag will use the same tag as the "main" image
        (defined by `image` attribute).
        only supported for docker or docker-buildx OCI-Builder. Must not be used in
        conjunction with `targets`.
        ''',
    ),
    AttributeSpec.optional(
        name='inputs',
        default={
            'repos': None, # None -> default to main repository
            'steps': {},
        },
        doc='configures the inputs that are made available to image build',
        type=dict, # todo: define types
    ),
    AttributeSpec.optional(
        name='prebuild_hook',
        default=None,
        doc='''
            if configured, a callback is executed prior to running image-build. the value is
            interpreted as relative path to main-repository root directory, and must be an
            executable file.
            It can be used, for example, to preprocess the "Dockerfile" to use, or to prepare
            contents within the build directory.
            The following environment variables are passed (all paths are absolute):
            - BUILD_DIR # path to build directory
            - DOCKERFILE # path to dockerfile

            Only supported for oci-builder docker or docker-buildx.
            dockerd will be available and running (docker excecutable accessible from PATH).
        ''',
        type=str,
    ),
    AttributeSpec.optional(
        name='tag_as_latest',
        default=False,
        doc='whether or not published container images should **also** be labeled as latest',
        type=bool,
    ),
    AttributeSpec.optional(
        name='tag_template',
        default='${EFFECTIVE_VERSION}',
        doc='the template to use for the image-tag (only variable: EFFECTIVE_VERSION)',
    ),
    AttributeSpec.optional(
        name='dockerfile',
        default='Dockerfile',
        doc='the file to use for building the container image',
    ),
    AttributeSpec.optional(
        name='dir',
        default=None,
        doc='the relative path to the container image build file',
    ),
    AttributeSpec.optional(
        name='target',
        default=None,
        doc='''\
            only for multistage builds: the target up to which to build.
            must not be used if `targets` is defined.
        ''',
    ),
    AttributeSpec.optional(
        name='targets',
        default=None,
        doc='''\
            if set, the given targets are built in the given order in the same build environment.
            This is useful to reduce resource consumption for multiple builds sharing common
            prerequisite build steps.
            Only supported if oci-builder is set to `docker-buildx` or `docker`
        ''',
        type=list[TargetSpec],
    ),
    AttributeSpec.optional(
        name='resource_labels',
        default=[],
        type=typing.List[ocm.Label],
        doc='labels to add to the resource declaration for this image in base-component-descriptor'
    ),
    AttributeSpec.optional(
        name='build_args',
        default={},
        type=typing.Dict[str, str],
        doc='build-time arguments to pass to docker-build',
    ),
    AttributeSpec.optional(
        name='platforms',
        doc=textwrap.dedent('''\
        If `platforms` is defined at toplevel, then defining it again for a single image-build
        can be done in order to only build this image for a subset of platforms.

        see toplevel documentation for `platforms` for reference.
        '''
        ),
        type=list[str],
        default=None,
    ),
)


class PublishDockerImageDescriptor(NamedModelElement, ModelDefaultsMixin, AttribSpecMixin):
    def __init__(
        self,
        *args,
        platform:str=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if not isinstance(self.raw, dict):
            raise ModelValidationError(
                f'{self.__class__.__name__} expects a dict - got: {self.raw=}'
            )
        self._apply_defaults(raw_dict=self.raw)
        self._platform = platform

        if platform:
            self._base_name = self._name
            self._name += f'-{platform.replace("/", "-")}'
        else:
            self._base_name = self._name

    @classmethod
    def _attribute_specs(cls):
        return IMG_DESCRIPTOR_ATTRIBS

    def _defaults_dict(self):
        return AttributeSpec.defaults_dict(IMG_DESCRIPTOR_ATTRIBS)

    def _optional_attributes(self):
        return set(AttributeSpec.optional_attr_names(IMG_DESCRIPTOR_ATTRIBS))

    def _required_attributes(self):
        return set(AttributeSpec.required_attr_names(IMG_DESCRIPTOR_ATTRIBS))

    def _inputs(self):
        return self.raw['inputs']

    def input_repos(self):
        return self._inputs()['repos']

    def input_steps(self):
        return self._inputs()['steps']

    def registry_name(self):
        return self.raw.get('registry')

    def image_reference(self):
        return self.raw['image']

    @property
    def extra_push_targets(self) -> list[om.OciImageReference]:
        return [om.OciImageReference(target) for target in self.raw['extra_push_targets']]

    def tag_as_latest(self) -> bool:
        return self.raw['tag_as_latest']

    def additional_tags(self) -> typing.Tuple[str]:
        if self.tag_as_latest():
            return ('latest',)
        return ()

    def tag_template(self):
        return self.raw['tag_template']

    def build_args(self):
        return self.raw['build_args']

    def target_name(self):
        return self.raw.get('target')

    @property
    def is_multitarget(self):
        if self.raw.get('targets'):
            return True
        return False

    @property
    def targets(self):
        if (target_name := self.target_name()):
            return (
                TargetSpec(
                    target=target_name,
                    image=self.image_reference(),
                    name=self._base_name,
                ),
            )
        if not (raw_targets := self.raw.get('targets')):
            return (
                TargetSpec(
                    target=None,
                    image=self.image_reference(),
                    name=self._base_name,
                ),
            )

        targets = []
        for raw_target in raw_targets:
            target = dacite.from_dict(
                data_class=TargetSpec,
                data=raw_target,
            )
            if not target.name:
                target.name = f'{self.name()}-{target.target}'

            targets.append(target)

        return targets

    @property
    def prebuild_hook(self) -> str | None:
        return self.raw['prebuild_hook']

    def dockerfile_relpath(self):
        return self.raw['dockerfile']

    def builddir_relpath(self):
        return self.raw['dir']

    def resource_labels(self):
        # for base-component-descriptor
        return self.raw['resource_labels']

    def resource_name(self):
        parts = self.image_reference().split('/')
        # image references are lengthy (e.g. gcr.eu/<org>/<path>/../<name>)
        # -> shorten this a bit (keep domain and last part of url path)
        domain = parts[0]
        image_name = parts[-1]
        return '_'.join([self.name(), domain, image_name])

    def platforms(self) -> list[str] | None:
        return self.raw['platforms']

    def platform(self) -> typing.Optional[str]:
        '''
        returns the target `platform` for which this image should be built, if tgt platform
        was explicitly configured in pipeline-definition.
        `None` indicates that no tgt platform was configured (in which case the platform
        should default to the runtime's own platform).
        '''
        return self._platform

    def _children(self):
        return ()

    def validate(self):
        super().validate()

        if self.target_name() and self.raw.get('targets', False):
            raise ModelValidationError('target and targets must not both be set')

        if self.extra_push_targets and self.raw.get('targets', False):
            raise ModelValidationError('targets and extra_push_targets must not both be set')

        for label in self.resource_labels():
            try:
                dacite.from_dict(
                    data_class=ocm.Label,
                    data=label,
                    config=dacite.Config(strict=True),
                )
            except dacite.DaciteError as e:
                raise ModelValidationError(
                    f"Invalid '{label=}'."
                ) from e


class OciBuilder(enum.Enum):
    DOCKER = 'docker'
    DOCKER_BUILDX = 'docker-buildx'


@dataclasses.dataclass
class HelmchartValueMapping:
    '''
    a mapping between OCM-Resources and values to be passed to a helmchart.

    Typically, this will be a reference to an OCI Image mapped to attributes within a
    "values.yaml" for helm.

    Mappings are evaluated as preprocessing step before creating
    helmchart-archive (i.e. values.yaml is patched). They are also incorporated into resulting
    OCM Component-Descriptors.

    ref: reference to (virtual) attribute of an OCM-Resource
    attribute: JSONPath-spec

    Syntax for ref:

        oci-resource:<name>.<attribute>
        ^^^^^
        prefix

        name: resource-name (match names used in `dockerimages` attribute)
        attribute: choose either of:
            repository: image-reference without tag
            tag:        tag (either symbolic tag, or digest)
            digest:     image-digest (including algorithm-prefix, e.g. sha256:...)
            image:      image-reference including tag

    Example:

    Assuming there is a component-descriptor with an OciImage-Resource named `my-image` that
    should be mapped to attributes `image.repository` and `image.tag` in values.yaml.

    - ref: ocm-resource:my-image.repository
      attribute: image.repository
    - ref: ocm-resource:my-image.tag
      attribute: image.tag
    '''
    ref: str
    attribute: str

    @property
    def referenced_resource_and_attribute(self):
        '''
        returns a two-tuple of reference resource-name and resource's attribute, parsed from
        `ref`. If ref is not a valid ref (see class-docstr), raises ValueError.
        '''
        if not self.ref.startswith(prefix := 'ocm-resource:'):
            raise ValueError(self.ref, f'must start with {prefix=}')

        ref = self.ref.removeprefix(prefix)
        parts = ref.split('.')
        if not len(parts) == 2:
            raise ValueError(ref, 'must consist of exactly two period-separated parts')

        resource_name, attribute_name = parts
        return resource_name, attribute_name


@dataclasses.dataclass
class HelmchartCfg:
    '''
    A configuration for a helmchart that should be published. Data is also injected into
    component-descriptor (if trait is present).

    name:     Chart's name (also used as resource-name for component-descriptor); must be unique
    registry: OCI Registry prefix to publish helmchart to (<name>:<version> will be appended)
    dir:      relative path to directory containing helmchart
    mappings: attributes to patch in values.yaml
    '''
    name: str
    registry: str
    dir: str
    mappings: list[HelmchartValueMapping] = dataclasses.field(default_factory=list)


ATTRIBUTES = (
    AttributeSpec.optional(
        name='dockerimages',
        doc='specifies the container images to be built',
        default={},
        type=typing.Dict[str, PublishDockerImageDescriptor],
    ),
    AttributeSpec.optional(
        name='oci-builder',
        doc='specifies the container image builder to use',
        type=OciBuilder,
        default=OciBuilder.DOCKER,
    ),
    AttributeSpec.optional(
        name='no-buildkit',
        doc='if using `docker` as oci-builder, force to not use buildkit - ignored otherwise',
        type=bool,
        default=False,
    ),
    AttributeSpec.optional(
        name='platforms',
        doc=textwrap.dedent('''\
        if defined, all image-builds will be done for each of the specified platforms, which
        may result in cross-platform builds. Only supported if using `docker-buildx` as oci-builder.

        As an implementation detail that may change in the future, multiarch/quemu-user-static
        `(see) <https://github.com/multiarch/qemu-user-static>`_ is used. As the underlying
        CICD nodes use `Linux` as kernel, only the architecture may be chosen. The following
        platforms are supported:

        - linux/386
        - linux/amd64 # aka x86_64
        - linux/arm/v6
        - linux/arm/v7
        - linux/arm64
        - linux/ppc64le
        - linux/riscv64
        - linux/s390x

        The resulting images will receive tags derived from the default tag (single-image case)
        with a suffix containing the platform.

        The default tag will be published as multi-arch image, referencing all platform-variants.
        This will also hold true if only one platform is specified.

        .. note::
            if specifying a list of platforms, _all_ platforms (including the default platform)
            must be explicitly specified.
        '''
        ),
        type=list[str],
        default=None,
    ),
    AttributeSpec.optional(
        name='helmcharts',
        doc='''\
        ''',
        type=tuple[HelmchartCfg],
        default=(),
    ),
)


class PublishTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def _children(self):
       return self.dockerimages()

    def dockerimages(self) -> typing.List[PublishDockerImageDescriptor]:
        image_dict = self.raw['dockerimages']

        if not (platforms := self.platforms()):
            platforms = (None, )

        def matches_platform(platform, image_args):
            if platform is None:
                return True

            if (platforms := image_args.get('platforms', None)) is None:
                return True

            if platform in platforms:
                return True

            # special-case: normalise x86_64 -> arm64
            if platform == 'linux/x86_64':
                return 'linux/amd64' in platforms
            if platform == 'linux/arm64':
                return 'linux/x86_64' in platforms

        for platform in platforms:
            yield from (
                PublishDockerImageDescriptor(
                    name,
                    args,
                    platform=platform,
                )
                for name, args
                in image_dict.items()
                if matches_platform(platform=platform, image_args=args)
            )

    def platforms(self) -> typing.Optional[list[str]]:
        '''
        the list of explicitly configured build platforms
        guaranteed to be either non-empty, or None
        '''
        return self.raw.get('platforms', None)

    def oci_builder(self) -> OciBuilder:
        return OciBuilder(self.raw['oci-builder'])

    def use_buildkit(self) -> bool:
        return not self.raw['no-buildkit']

    @property
    def helmcharts(self) -> tuple[HelmchartCfg]:
        return tuple(
            dacite.from_dict(
                HelmchartCfg,
                data=raw,
            )
            for raw in self.raw['helmcharts']
        )

    def transformer(self):
        return PublishTraitTransformer(trait=self)

    def validate(self):
        super().validate()

        if self.platforms() and not self.oci_builder() is OciBuilder.DOCKER_BUILDX:
            raise ModelValidationError(
                'must not specify platforms unless using docker-buildx as oci-builder'
            )
        if not self.platforms() and not self.platforms() is None:
            raise ModelValidationError(
                'must not specify empty list of platforms (omit attr instead)'
            )

        image_names = set()
        for image in self.dockerimages():
            if image.name() in image_names:
                raise ModelValidationError(
                    f'duplicate image: {image.name()=}',
                )
            image_names.add(image.name())

        helmchart_names = set()
        for helmchart in self.helmcharts:
            if helmchart.name in helmchart_names:
                raise ModelValidationError(
                    f'duplicate resource: {helmchart.name=}',
                )
            helmchart_names.add(helmchart.name)

            for mapping in helmchart.mappings:
                if not mapping.ref.startswith('ocm-resource:'):
                    raise ModelValidationError(
                        mapping,
                        'mapping.ref must start with ocm-resource:',
                    )

                try:
                    jsonpath_ng.parse(mapping.attribute)
                except jsonpath_ng.exceptions.JSONPathError as jpe:
                    raise ModelValidationError(
                        jpe,
                        mapping.attribute,
                    )

                _, attr = mapping.referenced_resource_and_attribute
                if not attr in (
                    'repository',
                    'tag',
                    'image',
                ):
                    raise ModelValidationError(
                        mapping.ref,
                        'mapping.ref\'s attribute must be one of repository, tag, image',
                    )


class PublishTraitTransformer(TraitTransformer):
    name = 'publish'

    def __init__(
        self,
        trait: PublishTrait,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.trait = not_none(trait)
        cfg_set = trait.cfg_set

        # XXX workaround for documentation-rendering
        if cfg_set:
            self.worker_node_cfg = cfg_set.concourse().worker_node_cfg
        self._build_steps = []

    def inject_steps(self):
        publish_step = PipelineStep(
            name='publish',
            raw_dict={},
            is_synthetic=True,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3,
            extra_args={
                'publish_trait': self.trait,
            },
        )
        publish_step.set_timeout(duration_string='4h')
        self._publish_step = publish_step

        # 'prepare' step
        prepare_step = PipelineStep(
            name='prepare',
            raw_dict={},
            is_synthetic=True,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.BOURNE_SHELL,
        )
        prepare_step.set_timeout(duration_string='30m')

        publish_step._add_dependency(prepare_step)

        if self.trait.helmcharts:
            helmcharts_step = PipelineStep(
                name='helmcharts',
                raw_dict={},
                is_synthetic=True,
                injecting_trait_name=self.name,
                script_type=ScriptType.PYTHON3,
                extra_args={
                    'publish_trait': self.trait,
                },
            )
            helmcharts_step._add_dependency(prepare_step)
            helmcharts_step.add_input(
                name=concourse.model.traits.component_descriptor.DIR_NAME,
                variable_name=concourse.model.traits.component_descriptor.ENV_VAR_NAME,
            )
            helmcharts_step.add_output(
                name='helmcharts',
                variable_name='helmcharts',
            )
            yield helmcharts_step

        if (oci_builder := self.trait.oci_builder()) in (
            OciBuilder.DOCKER,
            OciBuilder.DOCKER_BUILDX,
        ):
            if oci_builder in (OciBuilder.DOCKER, OciBuilder.DOCKER_BUILDX):
                pass
            else:
                raise NotImplementedError(oci_builder)

            for img in self.trait.dockerimages():
                worker_node_tags = ()

                if platform_name := img.platform():
                    platform = self.worker_node_cfg.platform_for_oci_platform(
                        oci_platform_name=platform_name,
                    )
                    if platform and platform.worker_tag:
                        worker_node_tags = (platform.worker_tag,)
                else:
                    platform = None

                build_step = PipelineStep(
                    name=f'build_oci_image_{img.name()}',
                    raw_dict={
                        'privilege_mode': PrivilegeMode.PRIVILEGED
                            if oci_builder in (OciBuilder.DOCKER, OciBuilder.DOCKER_BUILDX)
                            else PrivilegeMode.UNPRIVILEGED,
                    },
                    is_synthetic=True,
                    injecting_trait_name=self.name,
                    script_type=ScriptType.PYTHON3,
                    worker_node_tags=worker_node_tags,
                    platform=platform,
                    extra_args={
                        'image_descriptor': img,
                    }
                )
                build_step._add_dependency(prepare_step)
                self._build_steps.append(build_step)
                yield build_step

                publish_step._add_dependency(build_step)

        yield prepare_step
        yield publish_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        main_repo = pipeline_args.main_repository()
        prepare_step = pipeline_args.step('prepare')

        image_name = main_repo.branch() + '-image'
        tag_name = main_repo.branch() + '-tag'

        image_path = 'image_path'
        tag_path = 'tag_path'

        # configure prepare step's outputs (consumed by publish step)
        prepare_step.add_output(variable_name=image_path, name=image_name)
        prepare_step.add_output(variable_name=tag_path, name=tag_name)

        for build_step in self._build_steps + [self._publish_step]:
            build_step.add_input(variable_name=image_path, name=image_name)

        input_step_names = set()
        for image_descriptor in self.trait.dockerimages():
            # todo: image-specific prepare steps
            input_step_names.update(image_descriptor.input_steps())

        for input_step_name in input_step_names:
            input_step = pipeline_args.step(input_step_name)
            input_name = input_step.output_dir()
            prepare_step.add_input(input_name, input_name)

        # prepare-step depdends on every other step, except publish and release
        # TODO: do not hard-code knowledge about 'release' step
        for step in pipeline_args.steps():
            if step.name in ['publish', 'release', 'build_oci_image', 'helmcharts']:
                continue
            if step.name.startswith('build_oci_image'):
                continue
            if 'publish' in step.trait_depends():
                # don't depend on steps that have been explicitly configured to depend on publish.
                continue
            prepare_step._add_dependency(step)

    @classmethod
    def dependencies(cls):
        return {'version'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import unittest
from unittest.mock import MagicMock

import concourse.model.traits.component_descriptor as trait_module
from concourse.model.job import JobVariant


def component_descriptor_transformer(trait_dict: dict):
    trait = trait_module.ComponentDescriptorTrait(
        variant_name='does_not_matter',
        name='does_not_matter',
        raw_dict=trait_dict,
    )
    examinee = trait.transformer()
    return examinee


class ComponentDescriptorTraitTransformerTest(unittest.TestCase):
    def setUp(self):
        self.pipeline_args = JobVariant(
            name='a_job',
            raw_dict={},
            resource_registry=object(),
        )
        self.pipeline_args._steps_dict = {}
        self.pipeline_args._traits_dict = {}
        self.repo_mock = MagicMock()
        self.repo_mock.repo_hostname = MagicMock(return_value='github.com')
        self.repo_mock.repo_path = MagicMock(return_value='org/repo')
        self.pipeline_args.main_repository = MagicMock(return_value=self.repo_mock)

    def test_process_pipeline_args_injects_component_name(self):
        examinee = component_descriptor_transformer(trait_dict={})

        examinee.process_pipeline_args(pipeline_args=self.pipeline_args)

        self.assertEqual(
            examinee.trait.component_name(),
            'github.com/org/repo',
        )

    def test_process_pipeline_args_leaves_configured_component_name(self):
        examinee = component_descriptor_transformer(
            trait_dict={'component_name': 'foo.org/p/r'}
        )

        examinee.process_pipeline_args(pipeline_args=self.pipeline_args)

        self.assertEqual(
            examinee.trait.component_name(),
            'foo.org/p/r',
        )



import collections.abc
import dataclasses
import graphlib

import ci.util
import ocm
import oci.model as om

ComponentId = (
    ocm.Component
    | ocm.ComponentDescriptor
    | ocm.ComponentIdentity
    | ocm.ComponentReference
    | str
    | tuple[str, str]
)
ComponentName = (
    ocm.Component
    | ocm.ComponentReference
    | ocm.ComponentDescriptor
    | ocm.ComponentIdentity
    | str
)

META_SEPARATOR = '.build-'


def to_component_id(
    component: ComponentId, /
) -> ocm.ComponentIdentity:
    if isinstance(component, ocm.ComponentIdentity):
        return component

    if isinstance(component, ocm.ComponentDescriptor) or hasattr(component, 'component'):
        component = component.component
        # fall through to next case
    if isinstance(component, ocm.Component) or hasattr(component, 'name') \
        and not hasattr(component, 'componentName'):
        name = component.name
        version = component.version
    if isinstance(component, ocm.ComponentReference) or hasattr(component, 'componentName'):
        name = component.componentName
        version = component.version
    if isinstance(component, str):
        name, version = component.split(':', 1)
    if isinstance(component, tuple):
        name, version = component

    return ocm.ComponentIdentity(
        name=name,
        version=version,
    )


def to_component_name(
    component: ComponentName,
) -> str:
    if isinstance(component, ocm.ComponentDescriptor):
        component = component.component
    if isinstance(component, ocm.Component):
        component = component.name
    elif isinstance(component, ocm.ComponentIdentity):
        component = component.name
    elif isinstance(component, ocm.ComponentReference):
        component = component.componentName
    elif isinstance(component, tuple):
        if not len(component) == 2:
            raise ValueError('expected two-tuple with two elements')
        component = component[0]
    if not isinstance(component, str):
        raise ValueError(component)

    if ':' in component:
        # assumption: has form <name>:<version>
        # let exception raise in other cases
        component, _ = component.split(':')

    return component


def to_component_id_and_repository_url(
    component: ocm.Component | ocm.ComponentDescriptor | ocm.ComponentIdentity | str,
    repository: ocm.OciOcmRepository|str=None,
):
    if isinstance(component, str):
        name, version = component.rsplit(':', 1)
        component = ocm.ComponentIdentity(
            name=name,
            version=version,
        )

    if isinstance(component, ocm.ComponentDescriptor):
        component = component.component
    elif isinstance(component, ocm.ComponentIdentity) and not repository:
        raise ValueError('repository must be passed if calling w/ component-identity')

    if not repository: # component is sure to be of type ocm.Component by now (checked above)
        component: ocm.Component
        repository = component.current_ocm_repo

    if isinstance(repository, ocm.OciOcmRepository):
        repo_base_url = repository.baseUrl
    elif isinstance(repository, str):
        repo_base_url = repository
    else:
        raise ValueError(f'only OciOcmRepository is supported - got: {repository=}')

    return component, repo_base_url


def oci_ref(
    component: ocm.Component | ocm.ComponentDescriptor | ocm.ComponentIdentity | str,
    repository: ocm.OciOcmRepository|str=None,
) -> om.OciImageReference:
    component, repo_base_url = to_component_id_and_repository_url(
        component=component,
        repository=repository,
    )

    return om.OciImageReference(
        ci.util.urljoin(
            repo_base_url,
            'component-descriptors',
            f'{component.name.lower()}:{component.version}',
        )
    )


def iter_sorted(
    components: collections.abc.Iterable[ocm.Component], /
) -> collections.abc.Generator[ocm.Component, None, None]:
    '''
    returns a generator yielding the given components, honouring their dependencies, starting
    with "leaf" components (i.e. components w/o dependencies), also known as topologically sorted.
    '''
    components = (to_component(c) for c in components)
    components_by_id = {c.identity(): c for c in components}

    toposorter = graphlib.TopologicalSorter()

    def ref_to_comp_id(component_ref: ocm.ComponentReference) -> ocm.ComponentIdentity:
        return ocm.ComponentIdentity(
            name=component_ref.componentName,
            version=component_ref.version,
        )

    for component_id, component in components_by_id.items():
        depended_on_comp_ids = (
            ref_to_comp_id(cref)
            for cref in component.componentReferences
        )
        toposorter.add(component_id, *depended_on_comp_ids)

    for component_id in toposorter.static_order():
        if not component_id in components_by_id:
            # XXX: ignore component-references not contained in passed components for now
            continue

        yield components_by_id[component_id]


def to_component(*args, **kwargs) -> ocm.Component:
    if not kwargs and len(args) == 1:
        component = args[0]
        if isinstance(component, ocm.Component):
            return component
        elif isinstance(component, ocm.ComponentDescriptor):
            return component.component
        else:
            raise ValueError(args)
    raise NotImplementedError


def determine_component_name(
    repository_hostname: str,
    repository_path: str,
) -> str:
    component_name = '/'.join((
        repository_hostname,
        repository_path,
    ))
    return component_name.lower() # OCI demands lowercase


def normalise_component_name(component_name: str) -> str:
    return component_name.lower() # oci-spec demands lowercase


def oci_artefact_reference(
        component: (
            ocm.Component
            | ocm.ComponentIdentity
            | ocm.ComponentReference
            | str # 'name:version'
            | tuple[str, str] # (name, version)
        ),
        ocm_repository: str | ocm.OciOcmRepository = None
) -> str:
    if isinstance(component, ocm.Component):
        if not ocm_repository:
            ocm_repository = component.current_ocm_repo
        component_name = component.name
        component_version = component.version

    elif isinstance(component, ocm.ComponentIdentity):
        component_name = component.name
        component_version = component.version

    elif isinstance(component, ocm.ComponentReference):
        component_name = component.componentName
        component_version = component.version

    elif isinstance(component, str):
        component_name, component_version = component.split(':')

    elif isinstance(component, tuple):
        if not len(component) == 2 or not (
            isinstance(component[0], str) and isinstance(component[1], str)
        ):
            raise TypeError("If a tuple is given as component, it must contain two strings.")
        component_name, component_version = component
    else:
        # workaround: fallback to (deprecated) gci.componentmodel
        try:
            import gci.componentmodel
            if isinstance(component, gci.componentmodel.Component):
                if not ocm_repository:
                    ocm_repository = component.current_ocm_repo
                    ocm_repository = ocm.OciOcmRepository(
                        baseUrl=ocm_repository.baseUrl,
                    )
                component_name = component.name
                component_version = component.version
            else:
                raise TypeError(type(component))
        except ImportError:
            raise TypeError(type(component))

    if not ocm_repository:
        raise ValueError('ocm_repository must be given unless a Component is passed.')
    elif isinstance(ocm_repository, str):
        repo_ctx = ocm.OciOcmRepository(baseUrl=ocm_repository)
    elif isinstance(ocm_repository, ocm.OciOcmRepository):
        repo_ctx = ocm_repository
    else:
        raise TypeError(type(ocm_repository))

    return repo_ctx.component_version_oci_ref(
        name=component_name,
        version=component_version,
    )


def target_oci_ref(
    component: ocm.Component,
    component_ref: ocm.ComponentReference=None,
    component_version: str=None,
):
    if not component_ref:
        component_ref = component
        component_name = component_ref.name
    else:
        component_name = component_ref.componentName

    component_name = normalise_component_name(component_name)
    component_version = component_ref.version

    last_ctx_repo = component.current_ocm_repo

    return last_ctx_repo.component_version_oci_ref(
        name=component_name,
        version=component_version,
    )


def main_source(
    component: ocm.Component,
    absent_ok: bool=True,
) -> ocm.Source:
    component = to_component(component)
    for source in component.sources:
        if label := source.find_label('cloud.gardener/cicd/source'):
            if label.value.get('repository-classification') == 'main':
                return source

    if not component.sources:
        if absent_ok:
            return None
        raise ValueError(f'no sources defined by {component=}')

    # if no label was found use heuristic approach
    # heuristic: use first source
    return component.sources[0]


determine_main_source_for_component = main_source


@dataclasses.dataclass(frozen=True)
class ComponentResource:
    component: ocm.Component
    resource: ocm.Resource


@dataclasses.dataclass(frozen=True)
class LabelDiff:
    labels_only_left: list[ocm.Label] = dataclasses.field(default_factory=list)
    labels_only_right: list[ocm.Label] = dataclasses.field(default_factory=list)
    label_pairs_changed: list[tuple[ocm.Label, ocm.Label]] = dataclasses.field(default_factory=list)


empty_list = lambda: dataclasses.field(default_factory=list) # noqa:E3701


@dataclasses.dataclass
class ComponentDiff:
    cidentities_only_left: set = empty_list()
    cidentities_only_right: set = empty_list()
    cpairs_version_changed: list[tuple[ocm.Component, ocm.Component]] = empty_list
    # only set when new component is added/removed
    names_only_left: set = dataclasses.field(default_factory=set)
    names_only_right: set = dataclasses.field(default_factory=set)
    # only set on update
    names_version_changed: set = dataclasses.field(default_factory=set)


def diff_labels(
    left_labels: list[ocm.Label],
    right_labels: list[ocm.Label],
) -> LabelDiff:

    left_label_name_to_label = {l.name: l for l in left_labels}
    right_label_name_to_label = {l.name: l for l in right_labels}

    labels_only_left = [
        l for l in left_labels if l.name not in right_label_name_to_label.keys()
    ]
    labels_only_right = [
        l for l in right_labels if l.name not in left_label_name_to_label.keys()
    ]
    label_pairs_changed = []
    for left_label, right_label in _enumerate_group_pairs(
        left_elements=left_labels,
        right_elements=right_labels,
        unique_name=True,
    ):
        if left_label.value == right_label.value:
            continue
        else:
            label_pairs_changed.append((left_label, right_label))

    return LabelDiff(
        labels_only_left=labels_only_left,
        labels_only_right=labels_only_right,
        label_pairs_changed=label_pairs_changed,
    )


def diff_components(
    left_components: collections.abc.Iterable[ocm.Component],
    right_components: collections.abc.Iterable[ocm.Component],
    ignore_component_names=(),
) -> ComponentDiff:
    left_component_identities = {
        c.identity() for c in left_components if c.name not in ignore_component_names
    }
    right_component_identities = {
        c.identity() for c in right_components if c.name not in ignore_component_names
    }

    left_only_component_identities = left_component_identities - right_component_identities
    right_only_component_identities = right_component_identities - left_component_identities

    if left_only_component_identities == right_only_component_identities:
        return None # no diff

    left_components = tuple((
        c for c in left_components if c.identity() in left_only_component_identities
    ))
    right_components = tuple((
        c for c in right_components if c.identity() in right_only_component_identities
    ))

    def find_changed_component(
        changed_component: ocm.Component,
        components: list[ocm.Component],
    ):
        for c in components:
            if c.name == changed_component.name:
                return (changed_component, c)
        return (changed_component, None) # no pair component found

    components_with_changed_versions = []
    for component in left_components:
        changed_component = find_changed_component(component, right_components)
        if changed_component[1] is not None:
            components_with_changed_versions.append(changed_component)

    left_component_names = {i.name for i in left_component_identities}
    right_component_names = {i.name for i in right_component_identities}
    names_version_changed = {c[0].name for c in components_with_changed_versions}

    both_names = left_component_names & right_component_names
    left_component_names -= both_names
    right_component_names -= both_names

    return ComponentDiff(
        cidentities_only_left=left_only_component_identities,
        cidentities_only_right=right_only_component_identities,
        cpairs_version_changed=components_with_changed_versions,
        names_only_left=left_component_names,
        names_only_right=right_component_names,
        names_version_changed=names_version_changed,
    )


def _enumerate_group_pairs(
    left_elements: collections.abc.Sequence[ocm.Resource | ocm.Source | ocm.Label],
    right_elements: collections.abc.Sequence[ocm.Resource | ocm.Source, ocm.Label],
    unique_name: bool = False,
) -> collections.abc.Generator[tuple[list, list], None, None] | \
collections.abc.Generator[tuple, None, None]:
    '''Groups elements of two sequences with the same name.

    Can be used for Resources, Sources and Label
    '''
    # group the resources with the same name on both sides
    for element in left_elements:
        right_elements_group = [e for e in right_elements if e.name == element.name]
        # get resources for one resource via name

        # key is always in left group so we only have to check the length of the right group
        if len(right_elements_group) == 0:
            continue
        else:
            left_elements_group = [e for e in left_elements if e.name == element.name]

            if unique_name:
                if len(left_elements_group) == 1 and len(right_elements_group) == 1:
                    yield (left_elements_group[0], right_elements_group[0])
                else:
                    raise RuntimeError(
                        f'Element name "{element.name}"" is not unique at least one list. '
                        f'{len(left_elements_group)=} {len(right_elements_group)=}')
            else:
                yield (left_elements_group, right_elements_group)


@dataclasses.dataclass
class ResourceDiff:
    left_component: ocm.Component
    right_component: ocm.Component
    resource_refs_only_left: list[ocm.Resource] = dataclasses.field(default_factory=list)
    resource_refs_only_right: list[ocm.Resource] = dataclasses.field(default_factory=list)
    resourcepairs_version_changed: list[tuple[ocm.Resource, ocm.Resource]] = dataclasses.field(default_factory=list) # noqa:E501


def _add_if_not_duplicate(list, res):
    if (res.name, res.version) not in [(res.name, res.version) for res in list]:
        list.append(res)


def diff_resources(
    left_component: ocm.Component,
    right_component: ocm.Component,
) -> ResourceDiff:
    if type(left_component) is not ocm.Component:
        raise NotImplementedError(
            f'unsupported {type(left_component)=}',
        )
    if type(right_component) is not ocm.Component:
        raise NotImplementedError(
            f'unsupported {type(right_component)=}',
        )

    left_resource_identities_to_resource = {
        r.identity(left_component.resources + right_component.resources): r
        for r in left_component.resources
    }
    right_resource_identities_to_resource = {
        r.identity(left_component.resources + right_component.resources): r
        for r in right_component.resources
    }

    resource_diff = ResourceDiff(
        left_component=left_component,
        right_component=right_component,
    )

    if left_resource_identities_to_resource.keys() == right_resource_identities_to_resource.keys():
        return resource_diff

    left_names_to_resource = {r.name: r for r in left_component.resources}
    right_names_to_resource = {r.name: r for r in right_component.resources}
    # get left exclusive resources
    for resource in left_resource_identities_to_resource.values():
        if not resource.name in right_names_to_resource:
            _add_if_not_duplicate(resource_diff.resource_refs_only_left, resource)

    # get right exclusive resources
    for resource in right_resource_identities_to_resource.values():
        if not resource.name in left_names_to_resource:
            _add_if_not_duplicate(resource_diff.resource_refs_only_right, resource)

    # groups the resources by name. The version will be used at a later point
    def enumerate_group_pairs(
        left_resources: list[ocm.Resource],
        right_resources: list[ocm.Resource]
    ) -> collections.abc.Generator[tuple[list[ocm.Resource], list[ocm.Resource]], None, None]:
        # group the resources with the same name on both sides
        for name in left_names_to_resource.keys():
            right_resource_group = [r for r in right_resources if r.name == name]

            # key is always in left group so we only have to check the length of the right group
            if len(right_resource_group) == 0:
                continue
            else:
                left_resource_group = [r for r in left_resources if r.name == name]
                yield (left_resource_group, right_resource_group)

    for left_resource_group, right_resource_group in enumerate_group_pairs(
        left_resources=left_component.resources,
        right_resources=right_component.resources,
    ):
        if len(left_resource_group) == 1 and len(right_resource_group) == 1:
            # if versions are equal resource will be ignored, resource is unchanged
            if left_resource_group[0].version != right_resource_group[0].version:
                resource_diff.resourcepairs_version_changed.append(
                    (left_resource_group[0], right_resource_group[0]),
                )
            continue

        left_identities = {
            r.identity(left_component.resources + right_component.resources): r
            for r in left_resource_group
        }
        right_identities = {
            r.identity(left_component.resources + right_component.resources): r
                            for r in right_resource_group
        }

        left_resource_ids = sorted(left_identities.keys())
        right_resource_ids = sorted(right_identities.keys())

        # sort resources. Important because down/upgrades depend on position in list
        left_resources = [left_identities.get(id) for id in left_resource_ids]
        right_resources = [right_identities.get(id) for id in right_resource_ids]

        # remove all resources present in both
        versions_in_both = {
            r.version for r in left_resources
        } & {
            r.version for r in right_resources
        }
        left_resources = [
            i for i in left_resources
            if not i.version in versions_in_both
        ]
        right_resources = [
            i for i in right_resources
            if not i.version in versions_in_both
        ]

        # at this point we got left and right resources with the same name but different versions
        i = 0
        for i, left_resource in enumerate(left_resources):
            if i >= len(right_resources):
                _add_if_not_duplicate(resource_diff.resource_refs_only_left, left_resource)

            else:
                right_resource = right_resources[i]
                resource_diff.resourcepairs_version_changed.append((left_resource, right_resource))

        # returns an empyt dict if index out of bounds
        left_resource = left_resources[i:]
        right_resource = right_resources[i:]

        for i in left_resource:
            _add_if_not_duplicate(resource_diff.resource_refs_only_left, i)

        for i in right_resource:
            _add_if_not_duplicate(resource_diff.resource_refs_only_right, i)

    return resource_diff


def sanitise_version(version: str) -> str:
    '''
    Additional build metadata as defined in SemVer can be added via `+` to the version. However,
    OCI registries don't support `+` as tag character, which is why it has to be sanitised, for
    example using `META_SEPARATOR`.
    '''
    sanitised_version = version.replace('+', META_SEPARATOR)

    return sanitised_version


def desanitise_version(version: str) -> str:
    '''
    This function reverts the sanitisation of the `sanitise_version` function, which allows
    processing the version the same way as prior to using `sanitise_version`.
    '''
    desanitised_version = version.replace(META_SEPARATOR, '+')

    return desanitised_version



import logging

import awesomeversion

import delivery.model
import unixutil.model as um


logger = logging.getLogger(__name__)


def find_branch_info(
    os_id: um.OperatingSystemId,
    os_infos: list[delivery.model.OsReleaseInfo],
) -> delivery.model.OsReleaseInfo:
    if not os_id.ID:
        return None # os-id could not be determined

    os_version = os_id.VERSION_ID

    def version_candidates():
        yield os_version
        yield f'v{os_version}'

        parts = os_version.split('.')

        if len(parts) == 1:
            return

        yield parts[0]
        yield 'v' + parts[0]

        yield '.'.join(parts[:2]) # strip parts after minor
        yield 'v' + '.'.join(parts[:2]) # strip parts after minor

    candidates = tuple(version_candidates() if os_version else ())

    for os_info in os_infos:
        for candidate in candidates:
            if os_info.name == candidate:
                return os_info

    logger.warning(f'did not find branch-info for {os_id=}')


def branch_reached_eol(
    os_id: um.OperatingSystemId,
    os_infos: list[delivery.model.OsReleaseInfo],
) -> bool:
    branch_info = find_branch_info(
        os_id=os_id,
        os_infos=os_infos,
    )
    if not branch_info:
        return False

    return branch_info.reached_eol


def update_available(
    os_id: um.OperatingSystemId,
    os_infos: list[delivery.model.OsReleaseInfo],
    ignore_if_patchlevel_is_next_to_greatest=False,
) -> bool:
    branch_info = find_branch_info(
        os_id=os_id,
        os_infos=os_infos,
    )
    if not branch_info:
        return False

    if not branch_info.greatest_version:
        return False

    version = awesomeversion.AwesomeVersion(os_id.VERSION_ID.replace('_', '-'))
    greatest_version = awesomeversion.AwesomeVersion(branch_info.greatest_version.replace('_', '-'))

    greater_version_available = greatest_version > version

    if not greater_version_available or not ignore_if_patchlevel_is_next_to_greatest:
        return greater_version_available

    # there is greated version; check whether difference is not more than one patchlevel
    # check whether both versions actually _have_ patchlevel
    if not greatest_version.patch or not version.patch:
        return greater_version_available

    patch_diff = int(greatest_version.patch) - int(version.patch)
    return patch_diff > 1



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import typing

from abc import abstractmethod
from enum import Enum
from collections import namedtuple

import ci.util

from model.base import (
    ModelValidationError,
    ModelValidationMixin,
)


class EnumWithDocumentation(Enum):
    def __new__(cls, value, docstring=None):
        obj = object.__new__(cls)
        obj._value_ = value
        obj.__doc__ = docstring
        return obj

    def __str__(self):
        return str(self.value)


EnumValueWithDocumentation = namedtuple('EnumValueWithDocumentation', ['value', 'doc'])


class AttribSpecMixin:
    @classmethod
    def _attribute_specs(cls):
        raise NotImplementedError

    @classmethod
    def _defaults_dict(cls):
        return AttributeSpec.defaults_dict(cls._attribute_specs())

    @classmethod
    def _optional_attributes(cls):
        return set(AttributeSpec.optional_attr_names(cls._attribute_specs())) | \
            set(AttributeSpec.deprecated_attr_names(cls._attribute_specs()))

    @classmethod
    def _required_attributes(cls):
        return set(AttributeSpec.required_attr_names(cls._attribute_specs()))

    def _apply_defaults(self, raw_dict):
        self.raw = ci.util.merge_dicts(
            self._defaults_dict(),
            raw_dict,
        )


class ModelBase(AttribSpecMixin, ModelValidationMixin):
    def __init__(self, raw_dict: dict):
        ci.util.not_none(raw_dict)

        self._apply_defaults(raw_dict=raw_dict)
        self.custom_init(self.raw)

    def custom_init(self, raw_dict: dict):
        pass

    def _children(self):
        return ()


class RequiredPolicy(Enum):
    OPTIONAL = 'optional'
    REQUIRED = 'required'
    DEPRECATED = 'deprecated'


class AttributeSpec:
    @staticmethod
    def optional(name, doc, default, *args, **kwargs):
        return AttributeSpec(
            name=name,
            doc=doc,
            default=default,
            required=RequiredPolicy.OPTIONAL,
            *args,
            **kwargs,
        )

    @staticmethod
    def required(name, doc, *args, **kwargs):
        return AttributeSpec(
            name=name,
            doc=doc,
            required=RequiredPolicy.REQUIRED,
            *args,
            **kwargs,
        )

    @staticmethod
    def deprecated(name, doc, *args, **kwargs):
        return AttributeSpec(
            name=name,
            doc=doc,
            required=RequiredPolicy.DEPRECATED,
            *args,
            **kwargs,
        )

    @staticmethod
    def filter_attrs(attrs: 'typing.Iterable[AttributeSpec]', required: RequiredPolicy):
        if required:
            ci.util.check_type(required, RequiredPolicy)
        else:
            # no filtering
            yield from attrs

        for attr in attrs:
            ci.util.check_type(attr, AttributeSpec)
            if attr.required_policy() is required:
                yield attr

    @staticmethod
    def select_name(attr):
        ci.util.check_type(attr, AttributeSpec)
        return attr.name()

    @staticmethod
    def select_name_and_default(attr):
        ci.util.check_type(attr, AttributeSpec)
        return attr.name(), attr.default_value()

    @staticmethod
    def required_attr_names(attrs: 'typing.Iterable[AttributeSpec]'):
        yield from map(
            AttributeSpec.select_name,
            AttributeSpec.filter_attrs(attrs=attrs, required=RequiredPolicy.REQUIRED)
        )

    @staticmethod
    def optional_attr_names(attrs: 'typing.Iterable[AttributeSpec]'):
        yield from map(
            AttributeSpec.select_name,
            AttributeSpec.filter_attrs(attrs=attrs, required=RequiredPolicy.OPTIONAL)
        )

    @staticmethod
    def deprecated_attr_names(attrs: 'typing.Iterable[AttributeSpec]'):
        return {
            name: value for name, value in map(
                AttributeSpec.select_name_and_default,
                AttributeSpec.filter_attrs(attrs, required=RequiredPolicy.DEPRECATED,
                )
            )
        }

    @staticmethod
    def defaults_dict(attrs: 'typing.Iterable[AttributeSpec]'):
        return {
            name: value for name, value in map(
                AttributeSpec.select_name_and_default,
                AttributeSpec.filter_attrs(attrs, required=RequiredPolicy.OPTIONAL,
                )
            )
        }

    def __init__(
        self,
        name: str,
        doc: str,
        default=None,
        required=None,
        type=str,
    ):
        self._name = ci.util.check_type(name, str)
        self._doc = ci.util.check_type(doc, str)
        self._type = type

        # validate
        if default:
            if required and required not in (RequiredPolicy.OPTIONAL, RequiredPolicy.DEPRECATED):
                raise ValueError()

        self._required_policy = required
        self._default_value = default

    def name(self) -> str:
        return self._name

    def doc(self) -> str:
        return self._doc

    def default_value(self):
        return self._default_value

    def type(self) -> type:
        return self._type

    def required_policy(self):
        return self._required_policy

    def is_required(self):
        if self.required_policy() == RequiredPolicy.REQUIRED:
            return True
        elif self.required_policy() == RequiredPolicy.OPTIONAL:
            return False
        elif self.required_policy() == RequiredPolicy.DEPRECATED:
            return False
        raise NotImplementedError


class Trait(ModelBase):
    def __init__(
        self,
        name: str,
        variant_name: str,
        raw_dict: dict,
        cfg_set=None,
    ):
        self.name = ci.util.not_none(name)
        self.variant_name = ci.util.not_none(variant_name)
        self.cfg_set = cfg_set
        super().__init__(raw_dict=raw_dict)

    @abstractmethod
    def transformer(self):
        raise NotImplementedError

    def __str__(self):
        return 'Trait: {n}'.format(n=self.name)


class TraitTransformer:
    name = None # subclasses must overwrite

    def __init__(self):
        ci.util.not_none(self.name)

    def inject_steps(self):
        return []

    @classmethod
    def order_dependencies(cls):
        return set()

    @classmethod
    def dependencies(cls):
        return set()

    @abstractmethod
    def process_pipeline_args(self, pipeline_args):
        raise NotImplementedError()


class ScriptType(Enum):
    BOURNE_SHELL = 0
    PYTHON3 = 1


def normalise_to_dict(dictish):
    if type(dictish) == str:
        return {dictish: {}}
    if type(dictish) == list:
        values = []
        for v in dictish:
            if type(v) == dict and v:
                values.append(v.popitem())
            elif type(v) != dict:
                values.append((v, {}))
        return dict(values)
    return dictish


def fail(msg):
    raise ModelValidationError(msg)


def select_attr(name):
    return lambda o: o.name



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import collections.abc
import concurrent.futures
import dataclasses
import enum
import hashlib
import itertools
import json
import jsonschema
import logging
import os
import threading

import ccc.oci
import ci.util
import ctt.replicate
import cnudie.iter
import cnudie.retrieve
import cnudie.upload
import container.util
import dso.labels
import ocm
import oci
import oci.client
import oci.model as om

import ctt.filters as filters
import ctt.processing_model as processing_model
import ctt.processors as processors
import ctt.uploaders as uploaders
import ctt.util as ctt_util

original_tag_label_name = 'cloud.gardener.cnudie/migration/original_tag'

logger = logging.getLogger(__name__)

own_dir = os.path.abspath(os.path.dirname(__file__))


class ProcessingMode(enum.Enum):
    REGULAR = 'regular'
    DRY_RUN = 'dry_run'


class ProcessingPipeline:
    def __init__(
        self,
        name,
        filters,
        processor,
        uploaders,
    ):
        self._name = name
        self._filters = filters
        self._processor = processor
        self._uploaders = uploaders

    def matches(
        self,
        component: ocm.Component,
        resource: ocm.Resource,
    ):
        filters_count = len(self._filters)
        return all(
            map(
                lambda filtr, component, resource: filtr.matches(component, resource),
                self._filters,
                itertools.repeat(component, filters_count),
                itertools.repeat(resource, filters_count),
            )
        )

    def process(
        self,
        component: ocm.Component,
        resource: ocm.Resource,
        inject_ocm_coordinates_into_oci_manifests: bool=False,
    ) -> processing_model.ProcessingJob:
        if not self.matches(component, resource):
            return None

        logging.info(f'{inject_ocm_coordinates_into_oci_manifests=}')
        logging.info(
            f'{self._name} will process: '
            f'{component.name}:{resource.type}:{resource.access}'
        )

        job = processing_model.ProcessingJob(
            component=component,
            resource=resource,
            upload_request=processing_model.ContainerImageUploadRequest(
                source_ref=None,
                target_ref=None,  # must be set by a later step
                remove_files=None,  # _may_ be set by a later step
            ),
            inject_ocm_coordinates_into_oci_manifest=inject_ocm_coordinates_into_oci_manifests,
        )

        job: processing_model.ProcessingJob = self._processor.process(processing_job=job)

        first = True
        for uploader in self._uploaders:
            job = uploader.process(job, target_as_source=not first)
            first = False

        ctt_label = create_ctt_label(
            processing_rules=[
                self._name,
            ],
        )
        patched_resource = job.processed_resource.set_label(
            label=ctt_label,
        )
        job = dataclasses.replace(
            job,
            processed_resource=patched_resource,
        )

        return job


def create_ctt_label(
    processing_rules: list[str],
) -> ocm.Label:
    ctt_label_name = 'cloud.gardener/ctt-hint'
    label = ocm.Label(
        name=ctt_label_name,
        value={
            'processingRules': processing_rules,
        },
    )

    return label


def parse_processing_cfg(path: str):
    raw_cfg = ci.util.parse_yaml_file(path)

    processing_cfg_dir = os.path.abspath(os.path.dirname(path))
    for _, cfg in raw_cfg.get('processors', {}).items():
        cfg['kwargs']['base_dir'] = processing_cfg_dir

    return raw_cfg


def _filter(filter_cfg: dict):
    filter_ctor = getattr(filters, filter_cfg['type'])
    filter_ = filter_ctor(**filter_cfg.get('kwargs', {}))

    return filter_


def _processor(processor_cfg: dict):
    proc_type = processor_cfg['type']
    proc_ctor = getattr(processors, proc_type, None)
    if not proc_ctor:
        ci.util.fail(f'no such image processor: {proc_type}')
    processor = proc_ctor(**processor_cfg.get('kwargs', {}))
    return processor


def _uploader(uploader_cfg: dict):
    upload_type = uploader_cfg['type']
    upload_ctor = getattr(uploaders, upload_type, None)
    if not upload_ctor:
        ci.util.fail(f'no such uploader: {upload_type}')
    uploader = upload_ctor(**uploader_cfg.get('kwargs', {}))
    return uploader


def processing_pipeline(
    processing_cfg: dict,
    shared_processors: dict={},
    shared_uploaders: dict={},
) -> ProcessingPipeline:
    name = processing_cfg.get('name', '<no name>')

    filter_cfgs = processing_cfg['filter']
    if isinstance(filter_cfgs, dict):
        filter_cfgs = [filter_cfgs]
    filters = [_filter(filter_cfg=filter_cfg) for filter_cfg in filter_cfgs]

    if 'processor' in processing_cfg:
        processor_cfg = processing_cfg['processor']
        if isinstance(processor_cfg, str):
            proc = shared_processors[processor_cfg]
        else:
            proc = _processor(processor_cfg=processor_cfg)
    else:
        proc = processors.NoOpProcessor()

    upload_cfgs = processing_cfg['upload']
    if not isinstance(upload_cfgs, list):
        upload_cfgs = [upload_cfgs]  # normalise to list

    def instantiate_uploader(upload_cfg):
        if isinstance(upload_cfg, str):
            return shared_uploaders[upload_cfg]
        return _uploader(upload_cfg)

    uploaders = [instantiate_uploader(upload_cfg) for upload_cfg in upload_cfgs]

    pipeline = ProcessingPipeline(
        name=name,
        filters=filters,
        processor=proc,
        uploaders=uploaders,
    )
    return pipeline


def enum_processing_cfgs(
    processing_cfg: dict,
    shared_processors: dict,
    shared_uploaders: dict,
):
    cfg_entries = processing_cfg['image_processing_cfg']

    yield from map(
        processing_pipeline,
        cfg_entries,
        itertools.repeat(shared_processors, len(cfg_entries)),
        itertools.repeat(shared_uploaders, len(cfg_entries)),
    )


def determine_changed_components(
    component_descriptor: ocm.ComponentDescriptor,
    tgt_ocm_repo_url: str,
    component_descriptor_lookup: cnudie.retrieve.ComponentDescriptorLookupById,
    tgt_component_descriptor_lookup: cnudie.retrieve.ComponentDescriptorLookupById,
    component_filter: collections.abc.Callable[[ocm.Component], bool]=None,
    reftype_filter: collections.abc.Callable[[cnudie.iter.NodeReferenceType], bool]=None,
) -> collections.abc.Generator[ocm.ComponentDescriptor, None, None]:
    component = component_descriptor.component

    if component_filter and component_filter(component):
        return

    if tgt_component_descriptor_lookup(
        component.identity(),
        absent_ok=True,
    ):
        logger.info(
            f'{component.identity()} already exists in {tgt_ocm_repo_url=} '
            '- skipping replication of transitive closure'
        )
        return

    for cref in component.componentReferences:
        referenced_component_descriptor = component_descriptor_lookup(ocm.ComponentIdentity(
            name=cref.componentName,
            version=cref.version,
        ))

        yield from determine_changed_components(
            component_descriptor=referenced_component_descriptor,
            tgt_ocm_repo_url=tgt_ocm_repo_url,
            component_descriptor_lookup=component_descriptor_lookup,
            tgt_component_descriptor_lookup=tgt_component_descriptor_lookup,
            component_filter=component_filter,
            reftype_filter=reftype_filter,
        )

    if not (
        reftype_filter and reftype_filter(cnudie.iter.NodeReferenceType.EXTRA_COMPONENT_REFS_LABEL)
    ) and (
        label := component.find_label(dso.labels.ExtraComponentReferencesLabel.name)
    ):
        extra_crefs_label = dso.labels.deserialise_label(label)

        for extra_cref in extra_crefs_label.value:
            extra_cref_id = extra_cref.component_reference
            referenced_component_descriptor = component_descriptor_lookup(extra_cref_id)

            yield from determine_changed_components(
                component_descriptor=referenced_component_descriptor,
                tgt_ocm_repo_url=tgt_ocm_repo_url,
                component_descriptor_lookup=component_descriptor_lookup,
                tgt_component_descriptor_lookup=tgt_component_descriptor_lookup,
                component_filter=component_filter,
                reftype_filter=reftype_filter,
            )

    yield component_descriptor


def create_jobs(
    component_descriptors: collections.abc.Iterable[ocm.ComponentDescriptor],
    processing_cfg: dict,
    inject_ocm_coordinates: bool,
) -> collections.abc.Generator[processing_model.ProcessingJob, None, None]:
    shared_processors = {
        name: _processor(cfg) for name, cfg in processing_cfg.get('processors', {}).items()
    }
    shared_uploaders = {
        name: _uploader(cfg) for name, cfg in processing_cfg.get('uploaders', {}).items()
    }

    for component_descriptor in component_descriptors:
        component = component_descriptor.component

        for resource in component.resources:
            # XXX only support OCI-resources for now
            if not resource.access.type is ocm.AccessType.OCI_REGISTRY:
                continue

            for pipeline in enum_processing_cfgs(
                processing_cfg=processing_cfg,
                shared_processors=shared_processors,
                shared_uploaders=shared_uploaders,
            ):
                job = pipeline.process(
                    component=component,
                    resource=resource,
                    inject_ocm_coordinates_into_oci_manifests=inject_ocm_coordinates,
                )

                if not job:
                    continue # pipeline did not want to process

                yield job
                break
            else:
                ci.util.warning(
                    f'no matching processor: {component.name}:{resource.access}'
                )


uploaded_image_refs_to_digests = {}  # <ref>:<digest>
uploaded_image_refs_to_ready_events = {}  # <ref>:<event> (set if digest is available)
upload_image_lock = threading.Lock()


# uploads a single OCI artifact and returns the content digest
def process_upload_request(
    processing_job: processing_model.ProcessingJob,
    replication_mode: oci.ReplicationMode=oci.ReplicationMode.PREFER_MULTIARCH,
    platform_filter: collections.abc.Callable[[om.OciPlatform], bool]=None,
    oci_client: oci.client.Client=None,
) -> str:
    global uploaded_image_refs_to_digests
    global uploaded_image_refs_to_ready_events
    global upload_image_lock

    if not oci_client:
        oci_client = ccc.oci.oci_client()

    upload_request = processing_job.upload_request
    tgt_ref = upload_request.target_ref

    # if event is present, upload might still be in progress (done if event is "set")
    with upload_image_lock:
        if tgt_ref in uploaded_image_refs_to_ready_events:
            upload_done_event = uploaded_image_refs_to_ready_events[tgt_ref]
            wait_for_upload = True
        else:
            upload_done_event = threading.Event()
            uploaded_image_refs_to_ready_events[tgt_ref] = upload_done_event
            wait_for_upload = False

    if wait_for_upload:
        upload_done_event.wait()

    if tgt_ref in uploaded_image_refs_to_digests:  # digest already present
        logger.info(f'{tgt_ref=} - was already uploaded by another rule - skipping')
        return uploaded_image_refs_to_digests[tgt_ref]

    # most common case: tgt has not yet been processed - process and afterwards signal
    # other threads waiting for upload result that result is ready by setting the event

    accept = replication_mode.accept_header()
    manifest_blob_ref = oci_client.head_manifest(
        image_reference=tgt_ref,
        absent_ok=True,
        accept=accept,
    )
    if bool(manifest_blob_ref):
        logger.info(f'{tgt_ref=} exists - skipping upload')

        uploaded_image_refs_to_digests[tgt_ref] = manifest_blob_ref.digest
        upload_done_event.set()
        return manifest_blob_ref.digest

    src_ref = upload_request.source_ref

    logger.info(f'processing {src_ref} -> {tgt_ref=}')
    logger.info(f'{tgt_ref=} {upload_request.remove_files=} {replication_mode=} {platform_filter=}')

    component = processing_job.component
    resource = processing_job.resource

    if processing_job.inject_ocm_coordinates_into_oci_manifest:
        oci_manifest_annotations = {
            'cloud.gardener/ocm-component': f'{component.name}:{component.version}',
            'cloud.gardener/ocm-resource': f'{resource.name}:{resource.version}',
        }
    else:
        oci_manifest_annotations = None

    logging.info(f'{oci_manifest_annotations=}')

    try:
        _, _, raw_manifest = container.util.filter_image(
            source_ref=src_ref,
            target_ref=tgt_ref,
            remove_files=upload_request.remove_files,
            mode=replication_mode,
            platform_filter=platform_filter,
            oci_client=oci_client,
            oci_manifest_annotations=oci_manifest_annotations,
        )
    except Exception as e:
        logger.error(
            f'error trying to replicate {src_ref=} -> {tgt_ref=}'
        )
        e.add_note(f'filter_image: {src_ref=} -> {tgt_ref=}')
        raise e

    logger.info(f'finished processing {src_ref} -> {tgt_ref=}')

    manifest_digest = hashlib.sha256(raw_manifest).hexdigest()
    uploaded_image_refs_to_digests[tgt_ref] = f'sha256:{manifest_digest}'
    upload_done_event.set()
    return f'sha256:{manifest_digest}'


def process_images(
    processing_cfg_path: str,
    component_descriptor_v2: ocm.ComponentDescriptor,
    component_descriptor_lookup: cnudie.retrieve.ComponentDescriptorLookupById,
    processing_mode: ProcessingMode=ProcessingMode.REGULAR,
    replication_mode: oci.ReplicationMode=oci.ReplicationMode.PREFER_MULTIARCH,
    inject_ocm_coordinates_into_oci_manifests: bool=False,
    skip_cd_validation: bool=False,
    platform_filter: collections.abc.Callable[[om.OciPlatform], bool]=None,
    skip_component_upload: collections.abc.Callable[[ocm.Component], bool]=None,
    oci_client: oci.client.Client=None,
    component_filter: collections.abc.Callable[[ocm.Component], bool]=None,
    remove_label: collections.abc.Callable[[str], bool]=None,
    tgt_ocm_base_url: str=None,
    tgt_ctx_base_url: str=None, # deprecated -> replaced by `tgt_ocm_base_url`
) -> collections.abc.Generator[cnudie.iter.Node, None, None]:
    '''
    note: Passing a filter to prevent component descriptors from being replicated using the
    `skip_component_upload` parameter will still replicate all its resources (i.e. oci images)
    as well as referenced components. In contrast to that, passing a filter using the
    `component_filter` parameter will also exclude its resources as well as all transitive component
    references from the replication. In both cases, `True` means the respective component is
    _excluded_.
    '''
    if tgt_ctx_base_url:
        tgt_ocm_base_url = tgt_ctx_base_url

    if not tgt_ocm_base_url:
        raise ValueError(tgt_ocm_base_url)

    if not oci_client:
        oci_client = ccc.oci.oci_client()

    if processing_mode is ProcessingMode.DRY_RUN:
        ci.util.warning('dry-run: not downloading or uploading any images')

    src_ocm_base_url = component_descriptor_v2.component.current_ocm_repo.baseUrl

    if src_ocm_base_url == tgt_ocm_base_url:
        raise RuntimeError('current repo context and target repo context must be different!')

    executor = concurrent.futures.ThreadPoolExecutor(max_workers=16)

    reftype_filter = None
    if remove_label and remove_label(dso.labels.ExtraComponentReferencesLabel.name):
        def filter_extra_component_refs(reftype: cnudie.iter.NodeReferenceType) -> bool:
            return reftype is cnudie.iter.NodeReferenceType.EXTRA_COMPONENT_REFS_LABEL

        reftype_filter = filter_extra_component_refs

    tgt_component_descriptor_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
        ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(tgt_ocm_base_url),
        oci_client=oci_client,
        fallback_to_service_mapping=False,
    )

    component_descriptors = tuple(determine_changed_components(
        component_descriptor=component_descriptor_v2,
        tgt_ocm_repo_url=tgt_ocm_base_url,
        component_descriptor_lookup=component_descriptor_lookup,
        tgt_component_descriptor_lookup=tgt_component_descriptor_lookup,
        component_filter=component_filter,
        reftype_filter=reftype_filter,
    ))

    jobs = create_jobs(
        component_descriptors=component_descriptors,
        processing_cfg=parse_processing_cfg(processing_cfg_path),
        inject_ocm_coordinates=inject_ocm_coordinates_into_oci_manifests,
    )

    def process_job(processing_job: processing_model.ProcessingJob):
        if processing_mode is ProcessingMode.DRY_RUN:
            return processing_job
        elif processing_mode is ProcessingMode.REGULAR:
            pass
        else:
            raise NotImplementedError(processing_mode)

        oci_manifest_digest = process_upload_request(
            processing_job=processing_job,
            replication_mode=replication_mode,
            platform_filter=platform_filter,
            oci_client=oci_client,
        )

        if extra_tags := processing_job.extra_tags:
            target_ref = om.OciImageReference(processing_job.upload_request.target_ref)
            target_repo = target_ref.ref_without_tag
            manifest_bytes = oci_client.manifest_raw(
                image_reference=f'{target_repo}@{oci_manifest_digest}',
                accept=om.MimeTypes.prefer_multiarch,
            ).content

        for extra_tag in extra_tags:
            push_target = f'{target_repo}:{extra_tag}'

            oci_client.put_manifest(
                image_reference=push_target,
                manifest=manifest_bytes,
            )

        if not oci_manifest_digest:
            raise RuntimeError(f'No oci_manifest_digest returned for {processing_job=}')

        processed_resource = processing_job.processed_resource

        if processed_resource and (digest := processed_resource.digest):
            # if resource has a digest we understand, and is an ociArtifact, then we need to
            # update the digest, because we might have changed the oci-artefact
            if (
                digest.hashAlgorithm.upper() == 'SHA-256'
                and digest.normalisationAlgorithm == ocm.NormalisationAlgorithm.OCI_ARTIFACT_DIGEST
            ):
                digest.value = oci_manifest_digest.removeprefix('sha256:')

                processed_resource = dataclasses.replace(
                    processed_resource,
                    digest=digest,
                )

        if processing_job.upload_request.reference_target_by_digest:
            target_ref = om.OciImageReference.to_image_ref(processing_job.upload_request.target_ref)

            if (
                processing_job.upload_request.retain_symbolic_tag
                and (target_ref.has_symbolical_tag or target_ref.has_mixed_tag)
            ):
                target_ref = f'{target_ref.with_symbolical_tag}@{oci_manifest_digest}'
            else:
                target_ref = f'{target_ref.ref_without_tag}@{oci_manifest_digest}'

            access = processed_resource.access
            if access.type is ocm.AccessType.OCI_REGISTRY:
                access = dataclasses.replace(
                    processed_resource.access,
                    imageReference=target_ref,
                )

            elif access.type is ocm.AccessType.RELATIVE_OCI_REFERENCE:
                access = dataclasses.replace(
                    processed_resource.access,
                    reference=om.OciImageReference.to_image_ref(
                        image_reference=target_ref,
                        normalise=False, # don't inject docker special handlings
                    ).local_ref,
                )

            processed_resource = dataclasses.replace(
                processed_resource,
                access=access,
            )

            processing_job.processed_resource = processed_resource

            processing_job.upload_request = dataclasses.replace(
                processing_job.upload_request,
                target_ref=target_ref,
            )

        return processing_job

    def wrap_process_job(processing_job: processing_model.ProcessingJob):
        try:
            return process_job(processing_job=processing_job)
        except Exception as e:
            logger.error(f'exception while processing {processing_job=}')
            raise e

    jobs = tuple(executor.map(wrap_process_job, jobs))

    def append_ocm_repo(
        component: ocm.Component,
        ocm_repo: str | ocm.OciOcmRepository,
    ):
        if isinstance(ocm_repo, str):
            ocm_repo = ocm.OciOcmRepository(baseUrl=ocm_repo)
        elif isinstance(ocm_repo, ocm.OciOcmRepository):
            pass
        else:
            raise TypeError(ocm_repo)

        if component.current_ocm_repo.baseUrl != ocm_repo.baseUrl:
            component.repositoryContexts.append(ocm_repo)

    for component_descriptor in component_descriptors:
        component = component_descriptor.component

        job_group = [
            job for job in jobs
            if job.component.identity() == component.identity()
        ]

        patched_resources = {}

        # patch-in overwrites (caveat: must be done sequentially, as lists are not threadsafe)
        for job in job_group:
            patched_resource = job.processed_resource or job.resource
            patched_resources[job.resource.identity(component.resources)] = patched_resource

        component.resources = [
            patched_resources.get(resource.identity(component.resources), resource)
            for resource in component.resources
        ]

        append_ocm_repo(
            component=component,
            ocm_repo=tgt_ocm_base_url,
        )

        if remove_label:
            component.labels = [label for label in component.labels if not remove_label(label.name)]

        # Validate the patched component-descriptor and exit on fail
        if not skip_cd_validation:
            # ensure component-descriptor is json-serialisable
            raw = dataclasses.asdict(component_descriptor)
            try:
                raw_json = json.dumps(raw, cls=ctt_util.EnumJSONEncoder)
            except Exception as e:
                logger.error(f'Component-Descriptor could not be json-serialised: {e}')
                raise
            try:
                raw = json.loads(raw_json)
            except Exception as e:
                logger.error(f'Component-Descriptor could not be deserialised: {e}')
                raise

            try:
                ocm.ComponentDescriptor.validate(raw, validation_mode=ocm.ValidationMode.FAIL)
            except jsonschema.exceptions.RefResolutionError as rre:
                logger.warning(
                    'error whilst resolving reference from json-schema (see below) - will ignore'
                )
                print(rre)
            except Exception as e:
                component_id = f'{component.name}:{component.version}'
                logger.warning(
                    f'Schema validation for component-descriptor {component_id} failed with {e}'
                )

        # publish the (patched) component-descriptors
        if skip_component_upload and skip_component_upload(component):
            continue

        if processing_mode is ProcessingMode.DRY_RUN:
            print('dry-run - will not publish component-descriptor')
            continue
        elif processing_mode is not ProcessingMode.REGULAR:
            raise NotImplementedError(processing_mode)

        if len(ocm_repos := component.repositoryContexts) >= 2:
            orig_ocm_repo = component.repositoryContexts[-2]
        elif len(ocm_repos) == 1:
            logger.warning(f'{component.name}:{component.version} has only one ocm-repository')
            logger.warning('(expected: two or more)')
            logger.warning(f'{ocm_repos=}')
            orig_ocm_repo = component.repositoryContexts[-1]
        else:
            raise RuntimeError(f'{component.name}:{component.version} has no ocm-repository')

        ctt.replicate.replicate_oci_artifact_with_patched_component_descriptor(
            src_name=component.name,
            src_version=component.version,
            patched_component_descriptor=component_descriptor,
            src_ctx_repo=orig_ocm_repo,
        )

    # retrieve component descriptor from the target registry as local descriptor might not contain
    # patched image references (if it was already existing the the target registry and thus patching
    # has been skipped)
    if not skip_component_upload or not skip_component_upload(component_descriptor_v2.component):
        component_descriptor_v2 = tgt_component_descriptor_lookup(ocm.ComponentIdentity(
            name=component_descriptor_v2.component.name,
            version=component_descriptor_v2.component.version,
        ))

    for node in cnudie.iter.iter(
        component=component_descriptor_v2,
        lookup=tgt_component_descriptor_lookup,
        component_filter=component_filter,
        reftype_filter=reftype_filter,
    ):
        if cnudie.iter.Filter.components(node):
            pass
        elif cnudie.iter.Filter.resources(node):
            node: cnudie.iter.ResourceNode

            if node.resource.access.type not in (
                ocm.AccessType.OCI_REGISTRY,
                ocm.AccessType.RELATIVE_OCI_REFERENCE,
            ):
                continue
        else:
            continue

        yield node



import dataclasses
import datetime
import logging
import os
import typing

import pytimeparse
import ruamel.yaml
import ruamel.yaml.scalarstring

import cfg_mgmt.model as cmm
import cfg_mgmt.rotate as cmro
import concourse.util
import gitutil
import model
import model.base


logger = logging.getLogger(__name__)


def iter_cfg_elements(
    cfg_factory: typing.Union[model.ConfigFactory, model.ConfigurationSet],
    cfg_target: typing.Optional[cmm.CfgTarget] = None,
):
    if isinstance(cfg_factory, model.ConfigurationSet):
        type_names = cfg_factory.cfg_factory._cfg_types().keys()
    else:
        type_names = cfg_factory._cfg_types().keys()

    for type_name in type_names:
        # workaround: cfg-sets may reference non-local cfg-elements
        # also, cfg-elements only contain references to other cfg-elements
        # -> policy-checks will only add limited value
        if type_name == 'cfg_set':
            continue
        for cfg_element in cfg_factory._cfg_elements(cfg_type_name=type_name):
            if cfg_target and not cfg_target.matches(cfg_element):
                continue
            yield cfg_element


def iter_cfg_queue_entries_to_be_deleted(
    cfg_metadata: cmm.CfgMetadata,
    cfg_target: typing.Optional[cmm.CfgTarget]=None,
) -> typing.Generator[cmm.CfgQueueEntry, None, None]:
    now = datetime.datetime.now()
    for cfg_queue_entry in cfg_metadata.queue:
        if cfg_target and not cfg_target == cfg_queue_entry.target:
            continue

        if not cfg_queue_entry.to_be_deleted(now):
            continue

        yield cfg_queue_entry


def create_config_queue_entry(
    queue_entry_config_element: model.NamedModelElement,
    queue_entry_data: dict,
    delete_after_period: str,
) -> cmm.CfgQueueEntry:
    return cmm.CfgQueueEntry(
        target=cmm.CfgTarget(
            name=queue_entry_config_element.name(),
            type=queue_entry_config_element._type_name,
        ),
        deleteAfter=(
            datetime.datetime.today()
            + datetime.timedelta(seconds=pytimeparse.parse(delete_after_period))
        ).date().isoformat(),
        secretId=queue_entry_data,
    )


def update_config_status(
    cfg_status_file_path: str,
    config_element: model.NamedModelElement,
    config_statuses: typing.Iterable[cmm.CfgStatus],
):
    for cfg_status in config_statuses:
        if cfg_status.matches(
            element=config_element,
        ):
            break
    else:
        # does not exist
        cfg_status = cmm.CfgStatus(
            target=cmm.CfgTarget(
                type=config_element._type_name,
                name=config_element.name(),
            ),
            credential_update_timestamp=datetime.date.today().isoformat(),
        )
        config_statuses.append(cfg_status)
    cfg_status.credential_update_timestamp = datetime.date.today().isoformat()

    yaml = ruamel.yaml.YAML()
    with open(cfg_status_file_path, 'w') as f:
        yaml.dump(
            {
                'config_status': [
                    dataclasses.asdict(cfg_status)
                    for cfg_status in config_statuses
                ]
            },
            f,
        )


def write_config_queue(
    cfg_dir: str,
    cfg_metadata: cmm.CfgMetadata,
    queue_file_name: str=cmm.cfg_queue_fname,
):
    yaml = ruamel.yaml.YAML()
    with open(os.path.join(cfg_dir, queue_file_name), 'w') as queue_file:
        yaml.dump(
            {
                'rotation_queue': [
                    dataclasses.asdict(cfg_queue_entry)
                    for cfg_queue_entry in cfg_metadata.queue
                ],
            },
            queue_file,
        )


def local_cfg_type_sources(
    cfg_element: model.NamedModelElement,
    cfg_factory: typing.Union[model.ConfigFactory, model.ConfigurationSet],
) -> set[str]:
    cfg_type = cfg_factory._cfg_type(cfg_element._type_name)
    return {
        src.file for src in cfg_type.sources() if isinstance(src, model.LocalFileCfgSrc)
    }


def _local_cfg_file(
    cfg_element: model.NamedModelElement,
    cfg_factory: typing.Union[model.ConfigFactory, model.ConfigurationSet],
) -> str | None:
    '''Return the local source file for the given config element from the given factory, or None
    if it is not unambiguosly identifiable.
    '''
    local_cfg_files = local_cfg_type_sources(cfg_element, cfg_factory)

    if len(local_cfg_files) > 1:
        logger.warning("Config elements with more than one local source file are not supported")
        return None
    if not local_cfg_files:
        logger.warning(f"No local source file known for cfg type '{cfg_element._type_name}'")
        return None

    return local_cfg_files.pop()


def write_named_element(
    cfg_element,
    cfg_dir: str,
    cfg_file_name: str,
):
    yaml = ruamel.yaml.YAML()
    with open(os.path.join(cfg_dir, cfg_file_name)) as cfg_file:
        file_contents = yaml.load(cfg_file)

    # ruamel can preserve anchors, unless the entries that are anchored are _directly_ manipulated.
    # To avoid this for kubernetes configs, only update the part that holds the kubeconfig
    if cfg_element._type_name == 'kubernetes':
        file_contents[cfg_element.name()]['kubeconfig'] = cfg_element.raw['kubeconfig']

        if cfg_element.raw.get('bound_secret_name'):
            file_contents[cfg_element.name()]['bound_secret_name'] = \
                cfg_element.raw['bound_secret_name']

    # for all other cases, just replace the contents of the top-level element (this will resolve
    # anchors)
    else:
        file_contents[cfg_element.name()] = cfg_element.raw

    # use util function to convert strings to more readable block format
    ruamel.yaml.scalarstring.walk_tree(file_contents)

    with open(os.path.join(cfg_dir, cfg_file_name), 'wt') as cfg_file:
        yaml.dump(file_contents, cfg_file)


def write_changes_to_local_dir(
    cfg_element: model.NamedModelElement,
    secret_id: dict,
    delete_after_period: str,
    cfg_metadata: cmm.CfgMetadata,
    cfg_factory: model.ConfigFactory,
    cfg_dir: str,
):
    src_file = _local_cfg_file(cfg_element, cfg_factory)

    if not src_file:
        raise RuntimeError(
            f"Unable to determine local source file for cfg type '{cfg_element._type_name}. '"
            'See previous warnings for more details.'
        )

    write_named_element(cfg_element, cfg_dir, src_file)

    cfg_metadata.queue.append(
        create_config_queue_entry(
            queue_entry_config_element=cfg_element,
            queue_entry_data=secret_id,
            delete_after_period=delete_after_period,
        )
    )
    write_config_queue(
        cfg_dir=cfg_dir,
        cfg_metadata=cfg_metadata,
    )

    update_config_status(
        config_element=cfg_element,
        config_statuses=cfg_metadata.statuses,
        cfg_status_file_path=os.path.join(
            cfg_dir,
            cmm.cfg_status_fname,
        )
    )


def rotate_config_element_and_persist_in_cfg_repo(
    cfg_element: model.NamedModelElement,
    cfg_factory: model.ConfigFactory,
    cfg_metadata: cmm.CfgMetadata,
    cfg_dir: str,
    github_cfg,
    github_repo_path: str,
    target_ref: str = 'refs/heads/master',
) -> bool:
    '''Rotate the given config element and write it to the given cfg-repo, along with any additional
    config metadata created.

    Returns `True` if the rotation was successful and `False` if no rotation was performed (for
    example due to there being no rotation-function for the given type).
    '''
    git_helper = gitutil.GitHelper(
        repo=cfg_dir,
        github_cfg=github_cfg,
        github_repo_path=github_repo_path,
    )

    src_file = _local_cfg_file(cfg_element, cfg_factory)
    if not src_file:
        raise RuntimeError(
            f"Unable to determine local source file for cfg type '{cfg_element._type_name}. '"
            'See previous warnings for more details.'
        )

    if ret_vals := cmro.rotate_cfg_element(
        cfg_factory=cfg_factory,
        cfg_element=cfg_element,
    ):
        revert_function, secret_id, updated_elem = ret_vals
    else:
        return False

    status = determine_status(
        element=cfg_element,
        policies=cfg_metadata.policies,
        rules=cfg_metadata.rules,
        responsibles=cfg_metadata.responsibles,
        statuses=cfg_metadata.statuses,
    )

    build_url = None
    if concourse.util._running_on_ci():
        try:
            build_url = concourse.util.own_running_build_url(cfg_factory=cfg_factory)
        except Exception as e:
            logger.warning(
                'Unable to determine own job-url. Will not put it in commit-message if a '
                f'commit is created. Error: {e}'
            )

    try:
        write_changes_to_local_dir(
            cfg_element=updated_elem,
            cfg_factory=cfg_factory,
            secret_id=secret_id,
            cfg_metadata=cfg_metadata,
            cfg_dir=cfg_dir,
            delete_after_period=status.policy.delete_after_period,
        )

        commit_message = f'rotate secret for {cfg_element._type_name}/{cfg_element.name()}'
        if build_url:
            commit_message += f'\n\nJob url: {build_url}'

        git_helper.add_and_commit(
            message=commit_message,
        )
        git_helper.push('@', target_ref)
    except Exception as e:
        logger.warning(f'failed to push updated secret - reverting. Error: {e}')
        revert_function()
        git_helper.repo.git.reset('--hard', '@~')
        # intentionally do not return False here, as we would try another rotation in our pipeline
        # in that case.

    return True


def process_cfg_queue_and_persist_in_repo(
    cfg_element: model.NamedModelElement,
    cfg_factory: model.ConfigFactory,
    cfg_metadata: cmm.CfgMetadata,
    cfg_queue_entry: cmm.CfgQueueEntry,
    cfg_dir: str,
    github_cfg,
    github_repo_path: str,
    target_ref: str = 'refs/heads/master',
):
    '''Process the given config-queue entry by deleting the referenced credentials in the
    infrastructure and persist the updated config metadata in the given config-repository.

    Returns `True` if the config-entry was processed successfully and `False` if no processing
    has taken place.
    '''
    git_helper = gitutil.GitHelper(
        repo=cfg_dir,
        github_cfg=github_cfg,
        github_repo_path=github_repo_path,
    )

    updated_element, processing_successful = cmro.delete_expired_secret(
        cfg_element=cfg_element,
        cfg_queue_entry=cfg_queue_entry,
        cfg_factory=cfg_factory,
    )

    if (
        not processing_successful
        and not updated_element
    ):
        # Nothing happened, return early
        return False

    if updated_element:
        # processing the queue has lead to an update in the element in question. Write it back.
        src_file = _local_cfg_file(cfg_element, cfg_factory)
        if not src_file:
            raise RuntimeError(
                f"Unable to determine local source file for cfg type '{cfg_element._type_name}. '"
                'See previous warnings for more details.'
            )
        write_named_element(updated_element, cfg_dir, src_file)

    if processing_successful:
        cfg_metadata.queue.remove(cfg_queue_entry)
        write_config_queue(
            cfg_dir=cfg_dir,
            cfg_metadata=cfg_metadata,
        )

    build_url = None
    if concourse.util._running_on_ci():
        try:
            build_url = concourse.util.own_running_build_url(cfg_factory=cfg_factory)
        except Exception as e:
            logger.warning(
                'Unable to determine own job-url. Will not put it in commit-message if a '
                f'commit is created. Error: {e}'
            )

    try:
        commit_message = f'process config queue for {cfg_element._type_name}/{cfg_element.name()}'
        if build_url:
            commit_message += f'\n\nJob url: {build_url}'
        git_helper.add_and_commit(
            message=commit_message,
        )
        git_helper.push('@', target_ref)
    except:
        logger.warning('failed to push processed config queue - reverting')
        git_helper.repo.git.reset('--hard', '@~')

    return processing_successful


def determine_status(
    element: model.NamedModelElement,
    policies: list[cmm.CfgPolicy],
    rules: list[cmm.CfgRule],
    responsibles: list[cmm.CfgResponsibleMapping],
    statuses: list[cmm.CfgStatus],
    element_storage: str=None,
) -> cmm.CfgElementStatusReport:
    for rule in rules:
        if rule.matches(element=element):
            break
    else:
        rule = None # no rule was configured

    rule: typing.Optional[cmm.CfgRule]

    if rule:
        for policy in policies:
            if policy.name == rule.policy:
                break
        else:
            rule = None # inconsistent cfg: rule with specified name does not exist
    else:
        policy = None

    for responsible in responsibles:
        if responsible.matches(element=element):
            break
    else:
        responsible = None

    for status in statuses:
        if status.matches(element):
            break
    else:
        status = None

    return cmm.CfgElementStatusReport(
        element_storage=element_storage,
        element_type=element._type_name,
        element_name=element._name,
        policy=policy,
        rule=rule,
        status=status,
        responsible=responsible,
    )



import collections
import dataclasses
import enum
import typing

import dacite
import yaml

'''
utils/model classes for CVSS (https://www.first.org/cvss/user-guide)
'''


class CVESeverity(enum.IntEnum):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    CRITICAL = 4
    BLOCKER = 5

    def reduce(
        self,
        minimum_severity: int,
        severity_classes=1,
    ) -> 'CVESeverity':
        return CVESeverity(max(minimum_severity, self.value - severity_classes))

    @staticmethod
    def from_cve_score(score: float):
        if score <= 0:
            return CVESeverity.NONE
        if score < 4.0:
            return CVESeverity.LOW
        if score < 7.0:
            return CVESeverity.MEDIUM
        if score < 9.0:
            return CVESeverity.HIGH
        return CVESeverity.CRITICAL


class AccessVector(enum.Enum):
    NETWORK = 'N'
    ADJACENT = 'A'
    LOCAL = 'L'
    PYSICAL = 'P'


class AttackComplexity(enum.Enum):
    LOW = 'L'
    HIGH = 'H'


class PrivilegesRequired(enum.Enum):
    NONE = 'N'
    LOW = 'L'
    HIGH = 'H'


class UserInteraction(enum.Enum):
    NONE = 'N'
    REQUIRED = 'R'


class Scope(enum.Enum):
    UNCHANGED = 'U'
    CHANGED = 'C'


class Confidentiality(enum.Enum):
    NONE = 'N'
    LOW = 'L'
    HIGH = 'H'


class Integrity(enum.Enum):
    NONE = 'N'
    LOW = 'L'
    HIGH = 'H'


class Availability(enum.Enum):
    NONE = 'N'
    LOW = 'L'
    HIGH = 'H'


@dataclasses.dataclass
class CVSSV3:
    access_vector: AccessVector
    attack_complexity: AttackComplexity
    user_interaction: UserInteraction
    privileges_required: PrivilegesRequired
    scope: Scope
    confidentiality: Confidentiality
    integrity: Integrity
    availability: Availability

    @staticmethod
    def attr_name_from_CVSS(name: str):
        if name == 'AV':
            return 'access_vector'
        elif name == 'AC':
            return 'attack_complexity'
        elif name == 'UI':
            return 'user_interaction'
        elif name == 'C':
            return 'confidentiality'
        elif name == 'I':
            return 'integrity'
        elif name == 'A':
            return 'availability'
        elif name == 'PR':
            return 'privileges_required'
        else:
            raise ValueError(name)

    @staticmethod
    def parse(cvss: str) -> 'CVSSV3':
        parts = {e.split(':')[0]: e.split(':')[1] for e in cvss.split('/')}
        return CVSSV3(
            access_vector=AccessVector(parts['AV']),
            attack_complexity=AttackComplexity(parts['AC']),
            privileges_required=PrivilegesRequired(parts['PR']),
            user_interaction=UserInteraction(parts['UI']),
            scope=Scope(parts['S']),
            confidentiality=Confidentiality(parts['C']),
            integrity=Integrity(parts['I']),
            availability=Availability(parts['A']),
        )

    @staticmethod
    def from_dict(cvss: dict) -> 'CVSSV3':
        return dacite.from_dict(
            data_class=CVSSV3,
            data=cvss,
            config=dacite.Config(
                cast=[
                    AccessVector,
                    AttackComplexity,
                    UserInteraction,
                    PrivilegesRequired,
                    Scope,
                    Confidentiality,
                    Integrity,
                    Availability,
                ],
            ),
        )

    def __str__(self) -> str:
        return (
            f'AV:{self.access_vector.value}/AC:{self.attack_complexity.value}/'
            f'PR:{self.privileges_required.value}/UI:{self.user_interaction.value}/'
            f'S:{self.scope.value}/C:{self.confidentiality.value}/'
            f'I:{self.integrity.value}/A:{self.availability.value}'
        )


class NetworkExposure(enum.Enum):
    PRIVATE = 'private'
    PROTECTED = 'protected'
    PUBLIC = 'public'


class InteractingUserCategory(enum.Enum):
    GARDENER_OPERATOR = 'gardener-operator'
    END_USER = 'end-user'


class CVENoneLowHigh(enum.Enum):
    NONE = 'none'
    LOW = 'low'
    HIGH = 'high'


@dataclasses.dataclass
class CveCategorisation:
    network_exposure: typing.Optional[NetworkExposure]
    authentication_enforced: typing.Optional[bool]
    user_interaction: typing.Optional[InteractingUserCategory]
    confidentiality_requirement: typing.Optional[CVENoneLowHigh]
    integrity_requirement: typing.Optional[CVENoneLowHigh]
    availability_requirement: typing.Optional[CVENoneLowHigh]
    comment: typing.Optional[str]

    @staticmethod
    def from_dict(raw: dict):
        return dacite.from_dict(
            data_class=CveCategorisation,
            data=raw,
            config=dacite.Config(cast=(enum.Enum,)),
        )


class Rescore(enum.Enum):
    REDUCE = 'reduce'
    NOT_EXPLOITABLE = 'not-exploitable'
    NO_CHANGE = 'no-change'


@dataclasses.dataclass
class RescoringRule:
    '''
    a CVE rescoring rule intended to be used when re-scoring a CVE (see `CVSSV3` type) for an
    artefact that has a `CveCategorisation`.

    category_value is expected of form <CveCategorisation-attrname>:<value> (see CveCategorisation
    for allowed values)
    cve_values is expected as a list of <CVSSV3-attrname>:<value> entries (see CVSSV3 for allowed
    values)
    rescore indicates the rescoring that should be done, if the rule matches.

    A rescoring rule matches iff the artefact's categorisation-value exactly matches the rule's
    category_value attr AND the CVE's values are included in the rule's `cve_values`.

    CVE-Attributes for which a rule does not specify any value are not considered for matching.
    '''
    category_value: str
    cve_values: list[str]
    rescore: Rescore
    name: typing.Optional[str] = None

    @property
    def category_attr(self):
        return self.category_value.split(':')[0]

    @property
    def category_type(self):
        attr = self.category_attr
        annotations = typing.get_type_hints(CveCategorisation)

        if not attr in annotations:
            raise ValueError(f'invalid category-name: {attr=}')

        attr = annotations[attr]
        if typing.get_origin(attr) == typing.Union:
            args = [a for a in typing.get_args(attr) if not isinstance(None, a)]
            if len(args) != 1:
                # indicate programming error (CveCategorisation attrs must either be
                # simple types, or optionals.
                raise ValueError(f'only typing.Optional | Union with None is allowed {args=}')

            return args[0]

        return annotations[attr]

    @property
    def parsed_category_value(self):
        category_type = self.category_type
        value = self.category_value.split(':', 1)[-1]

        # special-case for bool
        if category_type == bool:
            parsed = yaml.safe_load(value)
            if not isinstance(parsed, bool):
                raise ValueError(f'failed to parse {value=} into boolean (using yaml.parsing)')
            return parsed

        return category_type(value)

    @property
    def parsed_cve_values(self) -> dict[str, set[object]]:
        attr_values = collections.defaultdict(set)

        for cve_value in self.cve_values:
            attr, value = cve_value.split(':', 1)
            attr = CVSSV3.attr_name_from_CVSS(attr)

            if not attr in (annotations := typing.get_type_hints(CVSSV3)):
                raise ValueError(f'{attr=} is not an allowed cve-attrname')

            attr_type = annotations[attr]
            attr_values[attr].add(attr_type(value))

        return attr_values

    def matches_cvss(self, cvss: CVSSV3 | dict) -> bool:
        '''
        returns a boolean indicating whether this rule matches the given CVSS.

        Only CVSS-Attributes that are specified by this rule w/ at least one value are checked.
        If more than one value for the same attribute is specified, matching will be assumed if
        the given CVSS's attr-value is contained in the rule's values.
        '''
        cve_values = self.parsed_cve_values
        if not type(cvss) is dict:
            cvss = dataclasses.asdict(cvss)
        for attr, value in cvss.items():
            if not attr in cve_values:
                continue

            # if cvss is of type dict, the values have to be compared to the values of the enums
            if not value in [
                v if isinstance(value, enum.Enum) else v.value
                for v in cve_values[attr]
            ]:
                return False

        return True

    def matches_categorisation(self, categorisation: CveCategorisation) -> bool:
        attr = self.category_attr
        value = self.parsed_category_value

        return value == getattr(categorisation, attr)


def rescoring_rules_from_dicts(rules: list[dict]) -> typing.Generator[RescoringRule, None, None]:
    '''
    deserialises rescoring rules. Each dict is expected to have the following form:

    category_value: <CveCategorisation-attr>:<value>
    name: <str> (optional)
    rules:
      - cve_values:
        - <CVSSV3-attr>: <value>
        rescore: <Rescore>
    '''
    for rule in rules:
        category_value = rule['category_value']
        name = rule.get('name')

        for subrule in rule['rules']:
            cve_values = subrule['cve_values']
            rescore = subrule['rescore']

            yield dacite.from_dict(
                data_class=RescoringRule,
                data={
                    'category_value': category_value,
                    'cve_values': cve_values,
                    'rescore': rescore,
                    'name': name,
                },
                config=dacite.Config(
                    cast=(enum.Enum, tuple),
                )
            )


def matching_rescore_rules(
    rescoring_rules: typing.Iterable[RescoringRule],
    categorisation: CveCategorisation,
    cvss: CVSSV3 | dict,
) -> typing.Generator[RescoringRule, None, None]:
    for rescoring_rule in rescoring_rules:
        if not rescoring_rule.matches_categorisation(categorisation):
            continue
        if not rescoring_rule.matches_cvss(cvss):
            continue

        yield rescoring_rule


def rescore(
    rescoring_rules: typing.Iterable[RescoringRule],
    severity: CVESeverity,
    minimum_severity: int=CVESeverity.NONE,
) -> CVESeverity:
    for rule in rescoring_rules:
        if rule.rescore is Rescore.NO_CHANGE:
            continue
        elif rule.rescore is Rescore.REDUCE:
            severity = severity.reduce(
                severity_classes=1,
                minimum_severity=minimum_severity,
            )
        elif rule.rescore is Rescore.NOT_EXPLOITABLE:
            return CVESeverity(minimum_severity)
        else:
            raise NotImplementedError(rule.rescore)

    return severity



import ocm

import pytest

import test_utils

import concourse.steps
import concourse.model.traits.update_component_deps as update_component_deps
import concourse.model.traits.component_descriptor as component_descriptor


@pytest.fixture
def render_step():
    return concourse.steps.step_def('update_component_deps')


@pytest.fixture
def job_variant():
    update_component_deps_trait = update_component_deps.UpdateComponentDependenciesTrait(
        name='update_component_dependencies',
        variant_name='don\'t_care',
        raw_dict={
            'set_dependency_version_script':'some_path',
            'upstream_component_name':'don\'t_care',
        },
    )

    component_descriptor_trait = component_descriptor.ComponentDescriptorTrait(
        name='component_descriptor',
        variant_name='don\'t_care',
        raw_dict={
            'component_name': 'github.com/org/repo_name',
        },
    )

    component_descriptor_trait.ctx_repository = lambda: ocm.OciOcmRepository(
        baseUrl='dummy-base-url',
    )

    main_repo = test_utils.repository()

    job_variant = test_utils.job(main_repo)
    job_variant._traits_dict = {
        'update_component_deps': update_component_deps_trait,
        'component_descriptor': component_descriptor_trait,
    }
    return job_variant


def test_render_and_compile(
    render_step,
    job_variant,
):
    # as a smoke-test, just try to render
    step_snippet = render_step(
        job_step=None,
        job_variant=job_variant,
        github_cfg_name=None,
        indent=0
    )

    # try to compile (-> basic syntax check)
    compile(step_snippet, 'update_component_deps.mako', 'exec')



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
)


class ElasticSearchConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def endpoints(self):
        return self.raw['endpoints']

    def credentials(self):
        return ElasticSearchCredentials(raw_dict=self.raw['credentials'])

    def _required_attributes(self):
        return ('endpoint_url', 'endpoints')

    def _optional_attributes(self):
        return ('credentials',)


class ElasticSearchCredentials(BasicCredentials):
    pass



import tabulate
import textwrap
import typing

import checkmarx.client
import checkmarx.model as model


def get_scan_info_table(
    scan_results: typing.Iterable[model.ScanResult],
    tablefmt: str = 'simple',
):
    scan_info_header = ('Scan ID', 'Artifact Name', 'Artifact Version', 'Scan State', 'Start', 'End')

    def started_on(scan_result: model.ScanResult):
        if scan_result.scan_response:
            return scan_result.scan_response.dateAndTime.startedOn
        else:
            return 'unknown'

    def ended_on(scan_result: model.ScanResult):
        if scan_result.scan_response:
            return scan_result.scan_response.dateAndTime.finishedOn
        else:
            return 'unkown'

    scan_info_data = (
        (
            scan_result.scan_response.id,
            scan_result.artifact_name,
            scan_result.scanned_element.source.version,
            scan_result.scan_response.status.name,
            started_on(scan_result),
            ended_on(scan_result),
        ) for scan_result in scan_results
    )
    scan_info = tabulate.tabulate(
        headers=scan_info_header,
        tabular_data=scan_info_data,
        tablefmt=tablefmt,
    )
    return scan_info


def get_scan_statistics_tables(
        scan_results: typing.Iterable[model.ScanResult],
        routes: checkmarx.client.CheckmarxRoutes,
        tablefmt: str = 'simple',
):
    def artifact_name(scan_result: model.ScanResult):
        if tablefmt == 'html':
            return f'''
            <a href="{routes.web_ui_scan_history(scan_id=scan_result.scan_response.id)}">
                {scan_result.artifact_name}
            </a>
            '''
        else:
            return scan_result.artifact_name

    def scan_severity(scan_result: model.ScanResult):
        if tablefmt == 'html':
            scan_id = scan_result.scan_response.id
            project_id = scan_result.project_id
            return f'''
            <a href="{routes.web_ui_scan_viewer(scan_id=scan_id, project_id=project_id)}">
                {scan_result.scan_response.scanRiskSeverity}
            </a>
            '''
        else:
            return scan_result.scan_response.scanRiskSeverity

    scan_statistics_header = (
        'Artifact Name',
        'Artifact Version',
        'Overall Severity',
        'High',
        'Medium',
        'Low',
        'Info',
    )

    scan_statistics_data = [
        (
            artifact_name(scan_result),
            scan_result.scanned_element.source.version,
            scan_severity(scan_result),
            scan_result.scan_statistic.highSeverity,
            scan_result.scan_statistic.mediumSeverity,
            scan_result.scan_statistic.lowSeverity,
            scan_result.scan_statistic.infoSeverity,
        ) for scan_result in scan_results
    ]

    # monkeypatch: disable html escaping
    tabulate.htmlescape = lambda x: x
    scan_statistics = tabulate.tabulate(
        headers=scan_statistics_header,
        tabular_data=sorted(scan_statistics_data, key=lambda x: x[1], reverse=True),
        tablefmt=tablefmt,
        colalign=('left', 'center', 'center', 'center', 'center', 'center')
    )

    return scan_statistics


def print_scan_result(
    scan_results: typing.Iterable[model.ScanResult],
    routes: checkmarx.client.CheckmarxRoutes,
):
    scan_info_table = get_scan_info_table(scan_results=scan_results, tablefmt='simple')
    scan_statistics_table = get_scan_statistics_tables(
        scan_results=scan_results,
        tablefmt='simple',
        routes=routes,
    )

    print(scan_info_table)
    print('\n')
    print(scan_statistics_table)


def _mail_disclaimer():
    return textwrap.dedent('''
        <div>
          <p>
          Note: you receive this E-Mail, because you were configured as a mail recipient
          (see .ci/pipeline_definitions)
          To remove yourself, search for your e-mail address in said file and remove it.
          </p>
        </div>
    ''')


def assemble_mail_body(
    scans_above_threshold: typing.Dict,
    scans_below_threshold: typing.Dict,
    failed_artifacts: typing.Dict,
    threshold: int,
    routes: checkmarx.client.CheckmarxRoutes,
):
    body_parts = [_mail_disclaimer()]

    if len(scans_above_threshold) > 0:
        above_threshold_text = textwrap.dedent(f'''
            <p>
              The following scan artifacts in checkmarx were found to
              contain critical vulnerabilities (applying threshold {threshold})
            </p>
        ''')
        scan_statistics_above_threshold = get_scan_statistics_tables(
            scan_results=scans_above_threshold,
            tablefmt='html',
            routes=routes,
        )
        body_parts.append(above_threshold_text + scan_statistics_above_threshold)

    if len(scans_below_threshold) > 0:
        below_threshold_text = textwrap.dedent('''
            <p>
              The following scan artifacts were found to be below the configured threshold:
            </p>
        ''')
        scan_statistics_below_threshold = get_scan_statistics_tables(
            scan_results=scans_below_threshold,
            tablefmt='html',
            routes=routes,
        )
        body_parts.append(below_threshold_text + scan_statistics_below_threshold)

    if len(failed_artifacts) > 0:
        failed_artifacts_str = ''.join((
            f'<li>{artifact_name}</li>' for artifact_name in failed_artifacts
        ))
        failed_artifacts_text = textwrap.dedent(
            f'''
                <p>
                  The following artifacts finished in an erroneous state:
                    <ul>{failed_artifacts_str})</ul>
                </p>
            '''
        )
        body_parts.append(failed_artifacts_text)

    return ''.join(body_parts)



import copy
import dataclasses
import datetime
import dateutil.parser
import json
import logging
import typing

import dacite
import aliyunsdkcore.client
from aliyunsdkram.request.v20150501.ListAccessKeysRequest import ListAccessKeysRequest
from aliyunsdkram.request.v20150501.CreateAccessKeyRequest import CreateAccessKeyRequest
from aliyunsdkram.request.v20150501.DeleteAccessKeyRequest import DeleteAccessKeyRequest
from aliyunsdkcore.acs_exception.exceptions import ServerException

import cfg_mgmt
import ci.log
import model
import model.aws
import model.alicloud

from cfg_mgmt.model import (
    CfgQueueEntry,
    ValidationError,
)


logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


@dataclasses.dataclass
class AlicloudAccessKey:
    AccessKeyId: str
    Status: str
    AccessKeySecret: str
    CreateDate: datetime.datetime

    def __post_init__(self):
        self.CreateDate = dateutil.parser.isoparse(self.CreateDate)


@dataclasses.dataclass
class AccessKeyMetadata:
    AccessKeyId: str
    Status: str
    CreateDate: datetime.datetime

    def __post_init__(self):
        self.CreateDate = dateutil.parser.isoparse(self.CreateDate)


@dataclasses.dataclass
class CreateAccessKeyResponse:
    AccessKey: AlicloudAccessKey
    RequestId: str


@dataclasses.dataclass
class AccessKeysEnumeration:
    AccessKey: typing.List[AccessKeyMetadata]


@dataclasses.dataclass
class ListAccessKeysResponse:
    RequestId: str
    AccessKeys: AccessKeysEnumeration


def rotate_cfg_element(
    cfg_element: model.alicloud.AlicloudConfig,
    cfg_factory: model.ConfigFactory,
) ->  typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:

    client = aliyunsdkcore.client.AcsClient(
        ak=cfg_element.access_key_id(),
        secret=cfg_element.access_key_secret(),
    )

    request = CreateAccessKeyRequest()
    try:
        response = client.do_action_with_exception(request)
    except ServerException as e:
        if e.get_http_status() == 403:
            logger.warning(
                'User is not allowed to create a new access key. Please make sure that the '
                'Account is configured to allow users to manage their own keys.'
            )
        raise

    response: CreateAccessKeyResponse = dacite.from_dict(
        data_class=CreateAccessKeyResponse,
        data=json.loads(response),
        config=dacite.Config(check_types=False),
    )
    access_key = response.AccessKey

    raw_cfg = copy.deepcopy(cfg_element.raw)
    new_element = model.alicloud.AlicloudConfig(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )
    new_element.raw['access_key_id'] = access_key.AccessKeyId
    new_element.raw['access_key_secret'] = access_key.AccessKeySecret

    def revert_function():
        request = DeleteAccessKeyRequest()
        request.set_UserAccessKeyId(access_key.AccessKeyId)
        client.do_action_with_exception(request)

    secret_id = {'accessKeyId': cfg_element.access_key_id()}

    return revert_function, secret_id, new_element


def delete_config_secret(
    cfg_element: model.alicloud.AlicloudConfig,
    cfg_factory: model.ConfigFactory,
    cfg_queue_entry: CfgQueueEntry,
) -> model.alicloud.AlicloudConfig | None:
    client = aliyunsdkcore.client.AcsClient(
        ak=cfg_element.access_key_id(),
        secret=cfg_element.access_key_secret(),
    )
    access_key_id = cfg_queue_entry.secretId['accessKeyId']

    request = DeleteAccessKeyRequest()
    request.set_UserAccessKeyId(access_key_id)

    try:
        client.do_action_with_exception(request)
    except ServerException as e:
        if e.get_http_status() == 403:
            logger.warning(
                'User is not allowed to delete its own access key. Please make sure that '
                'the Account is configured to allow users to manage their own keys.'
            )
        raise

    return None


def validate_for_rotation(
    cfg_element: model.alicloud.AlicloudConfig,
):
    access_key_id = cfg_element.access_key_id()
    client = aliyunsdkcore.client.AcsClient(
        ak=access_key_id,
        secret=cfg_element.access_key_secret(),
    )

    request = ListAccessKeysRequest()
    try:
        response = client.do_action_with_exception(request)
    except ServerException as e:
        if e.get_http_status() == 403:
            logger.warning(
                'User is not allowed to list its own access keys. Please make sure that the '
                'Account is configured to allow users to manage their own keys.'
            )
            raise

    response: ListAccessKeysResponse = dacite.from_dict(
        data_class=ListAccessKeysResponse,
        data=json.loads(response),
        config=dacite.Config(check_types=False),
    )

    access_keys = response.AccessKeys.AccessKey

    if len(access_keys) == 2:
        raise ValidationError(
            'There are already two keys present in Alicloud for Access Key '
            f"'{access_key_id}'. Will not attempt rotation."
        )



import datetime
import functools
import logging
import random
import time
import typing

import ccc.concourse
import ccc.elasticsearch
import ccc.github
import ccc.secrets_server
import ci.util
import concourse.client.api
import concourse.client.model
import concourse.enumerator
import concourse.replicator
import model
import model.concourse
import model.webhook_dispatcher
import whd.dispatcher

from github3.exceptions import NotFoundError

from .pipelines import validate_repository_pipelines
from github.util import GitHubRepositoryHelper

from concourse.client.util import (
    jobs_not_triggered,
    pin_resource_and_trigger_build,
    PinningFailedError,
    PinningUnnecessary,
)
from concourse.client.model import (
    ResourceType,
)
from .model import (
    PullRequestAction,
    PullRequestEvent,
)
import whd.metric
import whd.util


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def process_pr_event(
    concourse_clients: typing.Iterable[concourse.client.api.ConcourseApiBase],
    cfg_factory,
    cfg_set: model.ConfigurationSet,
    whd_cfg: model.webhook_dispatcher.WebhookDispatcherConfig,
    pr_event: PullRequestEvent,
    es_client: ccc.elasticsearch.ElasticSearchClient,
    dispatch_start_time: datetime.datetime,
):
    if not (github_helper := github_api_for_pr_event(pr_event, cfg_set)):
        logger.error(
                f'Unable to create github-api for PR #{pr_event.number()} of '
                f'repository {pr_event.repository().repository_path()}.'
            )
        return

    if (
        pr_modified_pipeline_definitions(pr_event, cfg_set)
        and pr_event.action() in [PullRequestAction.OPENED, PullRequestAction.SYNCHRONIZE]
    ):
        logger.info(f'Validating .ci/pipeline-definition for PR #{pr_event.number()}')
        validate_pipeline_definitions(
            cfg_set=cfg_set,
            cfg_factory=cfg_factory,
            whd_cfg=whd_cfg,
            pr_event=pr_event,
            github_helper=github_helper,
            tagging_label='ci/broken-pipeline-definition',
        )

    for concourse_api in concourse_clients:
        resources = list(
            matching_resources(
                concourse_api=concourse_api,
                event=pr_event,
            )
        )

        if len(resources) == 0:
            continue

        if pr_event.action() is PullRequestAction.LABELED:
            required_labels = {
                resource.source.get('label')
                for resource in resources if resource.source.get('label') is not None
            }
            if (l := pr_event.label()) in ['lgtm', 'reviewed/lgtm']:
                # special case for label set by gardener-robot/prow in reaction to "\lgtm".

                if l in required_labels:
                    # don't set the required labels if the required labels were set
                    continue

                if not set_pr_labels(pr_event, github_helper, cfg_set, resources):
                    logger.warning(
                        f'Unable to set required labels for PR #{pr_event.number()} for '
                        f'repository {pr_event.repository().repository_path()}. Will not trigger '
                        'resource check.'
                    )
            elif required_labels and not (l in required_labels):
                # Label that was set will not trigger any pr-job.
                logger.info(
                    f"Label '{l}' was set, but is not required for any job that builds "
                    f"PR #{pr_event.number()} for repository "
                    f"'{pr_event.repository().repository_path()}'. Will not trigger "
                    'resource check.'
                )
                continue
        if (
            pr_event.action() in [PullRequestAction.OPENED, PullRequestAction.SYNCHRONIZE]
            and not set_pr_labels(pr_event, github_helper, cfg_set, resources)
        ):
            logger.warning(
                f'Unable to set required labels for PR #{pr_event.number()} for '
                f'repository {pr_event.repository().repository_path()}. Will not trigger '
                'resource check.'
            )
            continue

        logger.info(f'triggering resource check for PR #{pr_event.number()}')
        whd.util.trigger_resource_check(
            concourse_api=concourse_api,
            resources=resources,
        )
        ensure_pr_resource_updates(
            cfg_set=cfg_set,
            concourse_api=concourse_api,
            pr_event=pr_event,
            resources=resources,
            es_client=es_client,
        )
        # Give concourse a chance to react
        time.sleep(random.randint(5,10))
        handle_untriggered_jobs(pr_event=pr_event, concourse_api=concourse_api)

    webhook_delivery_metric = whd.metric.WebhookDelivery.create(
        delivery_id=pr_event.delivery(),
        event_type='pull_request',
        repository=pr_event.repository().repository_path(),
        hostname=pr_event.hostname(),
        process_total_seconds=(datetime.datetime.now() - dispatch_start_time).total_seconds(),
    )
    if es_client:
        ccc.elasticsearch.metric_to_es(
            es_client=es_client,
            metric=webhook_delivery_metric,
            index_name=whd.metric.index_name(webhook_delivery_metric),
        )


def matching_resources(
    concourse_api: concourse.client.api.ConcourseApiBase,
    event: PullRequestEvent,
) -> typing.Generator[concourse.client.model.PipelineConfigResource, None, None]:

    resources_gen = concourse_api.pipeline_resources(
        concourse_api.pipelines(),
        resource_type=ResourceType.PULL_REQUEST,
    )

    for resource in resources_gen:
        resource: concourse.client.model.PipelineConfigResource

        ghs = resource.github_source()
        repository = event.repository()
        if not ghs.hostname() == repository.github_host():
            continue
        if not ghs.repo_path().lstrip('/') == repository.repository_path():
            continue

        yield resource


def github_api_for_pr_event(
    pr_event: PullRequestEvent,
    cfg_set: model.ConfigurationSet,
):
    repo = pr_event.repository()
    github_host = repo.github_host()
    repository_path = repo.repository_path()

    github_cfg = ccc.github.github_cfg_for_repo_url(
        repo_url=ci.util.urljoin(github_host, repository_path),
        cfg_factory=cfg_set,
    )
    owner, name = repository_path.split('/')

    try:
        github_helper = GitHubRepositoryHelper(
            owner=owner,
            name=name,
            github_cfg=github_cfg,
        )
    except NotFoundError:
        logger.warning(
            f"Unable to access repository '{repository_path}' on github '{github_host}'. "
            "Please make sure the repository exists and the technical user has the necessary "
            "permissions to access it."
        )
        return None

    return github_helper


def validate_pipeline_definitions(
    cfg_factory,
    cfg_set: model.ConfigurationSet,
    whd_cfg: model.webhook_dispatcher.WebhookDispatcherConfig,
    pr_event: PullRequestEvent,
    github_helper: GitHubRepositoryHelper,
    tagging_label: str,
):
    repo_url = pr_event.repository().repository_url()
    job_mapping_set = cfg_set.job_mapping()
    job_mapping = job_mapping_set.job_mapping_for_repo_url(repo_url, cfg_set)
    pr_number = pr_event.number()

    try:
        validate_repository_pipelines(
            repo_url=pr_event.head_repository().repository_url(),
            cfg_set=cfg_factory.cfg_set(job_mapping.replication_ctx_cfg_set()),
            whd_cfg=whd_cfg,
            branch=pr_event.head_ref(),
            job_mapping=job_mapping,
        )
    except concourse.replicator.PipelineValidationError as e:
        # If validation fails add a comment on the PR iff we haven't already commented, as
        # tracked by label
        logger.warning(
            f'Pipeline-definition in PR #{pr_number} of repository {repo_url} failed '
            'validation. Commenting on PR.'
        )
        github_helper.add_comment_to_pr(
            pull_request_number=pr_number,
            comment=(
                'This PR proposes changes that would break the pipeline definition:\n'
                f'```\n{e}\n```\n'
            ),
        )
        if tagging_label not in pr_event.label_names():
            github_helper.add_labels_to_pull_request(pr_number, tagging_label)
    else:
        # validation succeeded. Remove the label again, if it is currently set.
        if tagging_label in pr_event.label_names():
            logger.info(
                f'Pipeline-definition in PR #{pr_number} of repository {repo_url} passed '
                'validation again. Commenting on PR.'
            )
            github_helper.remove_label_from_pull_request(pr_number, tagging_label)
            github_helper.add_comment_to_pr(
                pull_request_number=pr_number,
                comment='The pipeline-definition has been fixed.',
            )


def _should_label(
    job_mapping: model.concourse.JobMapping,
    github_helper: GitHubRepositoryHelper,
    sender_login: str,
    owner: str,
) -> bool:
    trusted_teams = job_mapping.trusted_teams()
    if trusted_teams:
        if (
            any(
                github_helper.is_team_member(team_name=t, user_login=sender_login)
                for t in trusted_teams
            )
        ):
            return True
        else:
            return False

    elif github_helper.is_org_member(organization_name=owner, user_login=sender_login):
        return True

    return False


def set_pr_labels(
    pr_event: PullRequestEvent,
    github_helper: GitHubRepositoryHelper,
    cfg_set: model.ConfigurationSet,
    resources,
) -> bool:
    '''
    @ return True if the required label was set
    '''
    required_labels = {
        resource.source.get('label')
        for resource in resources if resource.source.get('label') is not None
    }
    if not required_labels:
        return True

    pr_number = pr_event.number()
    sender_login = pr_event.sender()['login']

    repo = pr_event.repository()
    repository_path = repo.repository_path()
    owner, _ = repository_path.split('/')
    job_mapping_set: model.concourse.JobMappingSet = cfg_set.job_mapping()
    job_mapping = job_mapping_set.job_mapping_for_repo_url(
        repo.repository_url(), cfg_set
    )

    if pr_event.action() is PullRequestAction.OPENED:
        if _should_label(job_mapping, github_helper, sender_login, owner):
            logger.info(
                f"New pull request by trusted member '{sender_login}' in "
                f"'{repository_path}' found. Setting required labels '{required_labels}'."
            )
            github_helper.add_labels_to_pull_request(pr_number, *required_labels)
            return True
        else:
            logger.debug(
                f"New pull request by member in '{repository_path}' found, but creator is not "
                f"member of '{owner}' - will not set required labels."
            )
            github_helper.add_comment_to_pr(
                pull_request_number=pr_number,
                comment=(
                    f"Thank you @{sender_login} for your contribution. Before I can start "
                    "building your PR, a member of the organization must set the required "
                    f"label(s) {required_labels}. Once started, you can check the build "
                    "status in the PR checks section below."
                )
            )
            return False
    elif pr_event.action() is PullRequestAction.SYNCHRONIZE:
        if _should_label(job_mapping, github_helper, sender_login, owner):
            logger.info(
                f"Update to pull request #{pr_number} by trusted member '{sender_login}' "
                f" in '{repository_path}' found. "
                f"Setting required labels '{required_labels}'."
            )
            github_helper.add_labels_to_pull_request(pr_number, *required_labels)
            return True
        else:
            logger.debug(
                f"Update to pull request #{pr_number} by '{sender_login}' "
                f" in '{repository_path}' found. Ignoring, since they are not an org member'."
            )
            return False
    elif pr_event.action() is PullRequestAction.LABELED:
        if (l := pr_event.label()) in ['lgtm', 'reviewed/lgtm']:
            logger.info(
                f"The label '{l}' was added to pull request #{pr_number} on "
                f"'{repository_path}' by '{sender_login}'. "
                f"Setting required labels '{required_labels}'."
            )
            github_helper.add_labels_to_pull_request(pr_number, *required_labels)
            return True
    return False


def ensure_pr_resource_updates(
    cfg_set,
    concourse_api,
    pr_event: PullRequestEvent,
    resources: typing.List[concourse.client.model.PipelineConfigResource],
    es_client: ccc.elasticsearch.ElasticSearchClient,
    retries=10,
    sleep_seconds=3,
):
    time.sleep(sleep_seconds)

    retries -= 1
    if retries < 0:
        try:
            log_outdated_resources(
                cfg_set=cfg_set,
                outdated_resources=resources,
            )
        # ignore logging errors
        except BaseException:
            pass

        outdated_resources_names = [r.name for r in resources]
        logger.info(f'could not update resources {outdated_resources_names} - giving up')
        if es_client:
            resource_update_failed_metric = whd.metric.WebhookResourceUpdateFailed.create(
                delivery_id=pr_event.delivery(),
                repository=pr_event.repository().repository_path(),
                hostname=pr_event.hostname(),
                event_type='pull_request',
                outdated_resources_names=outdated_resources_names,
                pr_id=pr_event.pr_id(),
                pr_action=pr_event.action().value,
            )
            ccc.elasticsearch.metric_to_es(
                es_client=es_client,
                metric=resource_update_failed_metric,
                index_name=whd.metric.index_name(resource_update_failed_metric),
            )
        return

    def resource_versions(resource):
        return concourse_api.resource_versions(
            pipeline_name=resource.pipeline_name(),
            resource_name=resource.name,
        )

    def is_up_to_date(resource, resource_versions) -> bool:
        # check if pr requires a label to be present
        require_label = resource.source.get('label')
        if require_label:
            if require_label not in pr_event.label_names():
                logger.info('skipping PR resource update (required label not present)')
                # regardless of whether or not the resource is up-to-date, it would not
                # be discovered by concourse's PR resource due to policy
                return True

        # assumption: PR resource is up-to-date if our PR-number is listed
        # XXX hard-code structure of concourse-PR-resource's version dict
        pr_numbers = map(lambda r: r.version()['pr'], resource_versions)

        return str(pr_event.number()) in pr_numbers

    # filter out all resources that are _not_ up-to-date (we only care about those).
    # Also keep resources that currently fail to check so that we keep retrying those
    outdated_resources = [
        resource for resource in resources
        if resource.failing_to_check()
        or not is_up_to_date(resource, resource_versions(resource))
    ]

    if not outdated_resources:
        logger.info('no outdated PR resources found')
        return # nothing to do

    logger.info(f'found {len(outdated_resources)} PR resource(s) that require being updated')
    whd.util.trigger_resource_check(concourse_api=concourse_api, resources=outdated_resources)
    logger.info(f'retriggered resource check will try again {retries} more times')

    ensure_pr_resource_updates(
        concourse_api=concourse_api,
        cfg_set=cfg_set,
        pr_event=pr_event,
        resources=outdated_resources,
        retries=retries,
        sleep_seconds=sleep_seconds*1.2,
        es_client=es_client,
    )


def handle_untriggered_jobs(
    pr_event: PullRequestEvent,
    concourse_api,
):
    for job, resource, resource_version in jobs_not_triggered(pr_event, concourse_api):
        logger.info(
            f'processing untriggered job {job.name=} of {resource.pipeline_name()=} '
            f'{resource.name=} {resource_version.version()=}. Triggered by '
            f'{pr_event.action()=} of {pr_event.delivery()=}'
        )
        try:
            pin_resource_and_trigger_build(
                job=job,
                resource=resource,
                resource_version=resource_version,
                concourse_api=concourse_api,
                retries=3,
            )
        except PinningUnnecessary as e:
            logger.info(e)
        except PinningFailedError as e:
            logger.warning(e)


def pr_modified_pipeline_definitions(
    pr_event: PullRequestEvent,
    cfg_set,
) -> bool:
    pr_number = pr_event.number()
    if not (github_helper := github_api_for_pr_event(pr_event=pr_event, cfg_set=cfg_set)):
        return False

    changed_files = (f.filename for f in github_helper.repository.pull_request(pr_number).files())

    return '.ci/pipeline_definitions' in changed_files


@functools.lru_cache()
def els_client(
    cfg_set,
):
    elastic_cfg = cfg_set.elasticsearch()
    elastic_client = ccc.elasticsearch.from_cfg(elasticsearch_cfg=elastic_cfg)
    return elastic_client


def log_outdated_resources(
    cfg_set,
    outdated_resources,
):
    els_index = cfg_set.webhook_dispatcher_deployment().logging_els_index()
    elastic_client = els_client(cfg_set=cfg_set)

    date = datetime.datetime.utcnow().isoformat()
    elastic_client.store_documents(
        index=els_index,
        body=[
            {
                'date': date,
                'resource_name': resource.name,
                'pipeline_name': resource.pipeline_name(),
            }
            for resource in outdated_resources
        ],
    )



import datetime
import hashlib
import logging
import re
import textwrap
import traceback
import typing

import concourse.util
import github3
import github3.issues
import github3.issues.comment
import github3.issues.issue
import github3.issues.milestone
import github3.repos

import cfg_mgmt.model as cmm
import ci.log
import delivery.model
import github.compliance.model as gcm
import github.retry
import github.util

'''
functionality for creating and maintaining github-issues for tracking compliance issues
'''

logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


_label_checkmarx = 'vulnerabilities/checkmarx'
_label_bdba = 'vulnerabilities/bdba'
_label_licenses = 'licenses/bdba'
_label_os_outdated = 'os/outdated'
_label_malware = 'malware/clamav'
_label_diki = 'compliance/diki'

_label_no_responsible = 'cfg/policy-violation/no-responsible'
_label_no_rule = 'cfg/policy-violation/no-rule'
_label_no_status = 'cfg/policy-violation/no-status'
_label_outdated_credentials = 'cfg/policy-violation/credentials-outdated'
_label_undefined_policy = 'cfg/policy-violation/undefined-policy'

_label_prefix_ocm_artefact = 'ocm/artefact'
_label_prefix_cicd_cfg_element = 'cicd-cfg-element'
_label_prefix_ctx = 'ctx'
_label_overdue = 'overdue'
_label_scan_pending = 'scan-pending'


def prefix_for_element(
    scanned_element: gcm.Target,
) -> str:
    if gcm.is_ocm_artefact_node(scanned_element):
        return _label_prefix_ocm_artefact

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        return _label_prefix_cicd_cfg_element

    else:
        raise TypeError(scanned_element)


def unique_name_for_element(
    scanned_element: gcm.Target,
) -> str:
    if gcm.is_ocm_artefact_node(scanned_element):
        artifact = gcm.artifact_from_node(scanned_element)
        return f'{scanned_element.component.name}:{artifact.name}'

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        return scanned_element.name

    else:
        raise TypeError(scanned_element)


def digest_label(
    prefix: str,
    digest_str: str,
    max_length: int=50,
) -> str:
    '''
    concatenates and returns a fixed prefix with digest calculated from `digest_str`.

    this is useful as GitHub labels are limited to 50 characters
    '''
    digest_length = max_length - (len(prefix) + 1) # prefix + slash
    digest_length = int(digest_length / 2) # hexdigest is of double length

    # pylint does not know `length` parameter (it is even required, though!)
    digest = hashlib.shake_128(digest_str.encode('utf-8')).hexdigest( # noqa: E1123
        length=digest_length,
    )

    label = f'{prefix}/{digest}'

    if len(label) > max_length:
        raise ValueError(f'{digest_str=} and {prefix=} would result '
            f'in label exceeding length of {max_length=}')

    return label


def _search_labels(
    issue_type: str,
    scanned_element: gcm.Target=None,
    extra_labels: typing.Iterable[str]=(),
) -> typing.Generator[str, None, None]:
    if not issue_type:
        raise ValueError('issue_type must not be None or empty')

    if extra_labels:
        yield from extra_labels

    yield 'area/security'
    yield 'cicd/auto-generated'
    yield f'cicd/{issue_type}'

    if scanned_element:
        yield digest_label(
            prefix=prefix_for_element(
                scanned_element=scanned_element,
            ),
            digest_str=unique_name_for_element(
                scanned_element=scanned_element,
            ),
        )


@github.retry.retry_and_throttle
def enumerate_issues(
    known_issues: typing.Sequence[github3.issues.issue.ShortIssue],
    issue_type: str,
    scanned_element: gcm.Target=None,
    extra_labels: typing.Iterable[str]=(),
    state: str | None = None, # 'open' | 'closed ; `None` ignores issue state
) -> typing.Generator[github3.issues.ShortIssue, None, None]:
    '''Return an iterator iterating over those issues from `known_issues` that match the given
    parameters.
    '''
    labels = frozenset(_search_labels(
        scanned_element=scanned_element,
        issue_type=issue_type,
        extra_labels=extra_labels,
    ))

    def filter_relevant_issues(issue: github3.issues.issue.ShortIssue):
        # state is of type str, explicitly check for None
        if state != None and issue.state != state:
            return False

        issue_labels = frozenset((l.name for l in issue.original_labels))
        if not issue_labels & labels == labels:
            return False
        return True

    for issue in filter(filter_relevant_issues, known_issues):
        yield issue


@github.retry.retry_and_throttle
def _create_issue(
    issue_type: str,
    repository: github3.repos.Repository,
    body: str,
    title: str,
    scanned_element: gcm.Target=None,
    extra_labels: typing.Iterable[str]=(),
    assignees: typing.Iterable[str]=(),
    assignees_statuses: set[delivery.model.Status] = set(),
    milestone: github3.issues.milestone.Milestone=None,
    failed_milestones: list[github3.issues.milestone.Milestone]=[],
    latest_processing_date: datetime.date|datetime.datetime=None,
) -> github3.issues.issue.ShortIssue:
    assignees = tuple(assignees)

    labels = frozenset(_search_labels(
        scanned_element=scanned_element,
        issue_type=issue_type,
        extra_labels=extra_labels,
    ))

    try:
        issue = repository.create_issue(
            title=title,
            body=body,
            assignees=assignees,
            milestone=milestone.number if milestone else None,
            labels=sorted(labels),
        )

        if latest_processing_date:
            latest_processing_date = latest_processing_date.isoformat()
            issue.create_comment(f'{latest_processing_date=}')

        if assignees_statuses:
            comment_body = textwrap.dedent('''\
                There have been anomalies during initial ticket assignment, please see details below:
                | Message Type | Message |
                | --- | --- |'''
            )
            for status in assignees_statuses:
                comment_body += f'\n|`{status.type}`|`{status.msg}`'

            issue.create_comment(comment_body)

        if failed_milestones:
            milestone_comment = (
                'Failed to automatically assign ticket to one of these milestones: '
                f'{", ".join([milestone.title for milestone in failed_milestones])}. '
                'Milestones were probably closed before all associated findings were assessed. '
            )
            if milestone:
                milestone_comment += f'Ticket was assigned to {milestone.title} as a fallback.'
            issue.create_comment(milestone_comment)

        return issue

    except github3.exceptions.GitHubError as ghe:
        logger.warning(f'received error trying to create issue: {ghe=}')
        logger.warning(f'{ghe.message=} {ghe.code=} {ghe.errors=}')
        logger.warning(f'{labels=} {assignees=}')
        raise ghe


@github.retry.retry_and_throttle
def _update_issue(
    scanned_element: gcm.Target,
    issue_type: str,
    body:str,
    title:typing.Optional[str],
    issue: github3.issues.Issue,
    extra_labels: typing.Iterable[str]=(),
    milestone: github3.issues.milestone.Milestone=None,
    assignees: typing.Iterable[str]=(),
) -> github3.issues.issue.ShortIssue:
    kwargs = {}
    if not issue.assignees and assignees:
        kwargs['assignees'] = tuple(assignees)

    if title:
        kwargs['title'] = title

    if milestone and not issue.milestone:
        kwargs['milestone'] = milestone.number

    labels = sorted(_search_labels(
        scanned_element=scanned_element,
        issue_type=issue_type,
        extra_labels=extra_labels,
    ))

    kwargs['labels'] = labels

    issue.edit(
        body=body,
        **kwargs,
    )

    return issue


def _create_or_update_problems_comment(
    issue: github3.github.issues.ShortIssue,
    message: str,
):
    '''
    Create or update a comment on an issue stating encountered problems.
    '''

    job_url = concourse.util.own_running_build_url()
    now = datetime.datetime.now()

    issue_problems_comment_header = (
        'An error occurred when updating this issue:'
    )

    def find_existing_comment() -> github3.issues.comment.IssueComment | None:
        for comment in issue.comments(number=32):
            if issue_problems_comment_header in comment.body:
                return comment

    problems_comment = find_existing_comment()

    comment_body = (
        f'{issue_problems_comment_header}\n\n'
        f'{message}\n\n'
        f'Job url: {job_url}\n'
        f'Date of occurrence: `{now.isoformat(timespec="minutes")}`'
    )

    if not problems_comment:
        # Create new comment containing the message
        issue.create_comment(comment_body)
    else:
        # Update comment replacing old message
        problems_comment.edit(comment_body)


def create_or_update_issue(
    scanned_element: gcm.Target,
    issue_type: str,
    repository: github3.repos.Repository,
    body: str,
    known_issues: typing.Iterable[github3.issues.issue.ShortIssue],
    title: str,
    assignees: typing.Iterable[str]=(),
    assignees_statuses: set[delivery.model.Status] = set(),
    milestone: github3.issues.milestone.Milestone=None,
    failed_milestones: list[github3.issues.milestone.Milestone]=[],
    latest_processing_date: datetime.date=None,
    extra_labels: typing.Iterable[str]=None,
    preserve_labels_regexes: typing.Iterable[str]=(),
    ctx_labels: typing.Iterable[str]=(),
) -> github3.issues.issue.ShortIssue:
    '''
    Creates or updates a github issue for the given scanned_element. If no issue exists, yet, it will
    be created, otherwise, it is checked that at most one matching issue exists; which will then be
    updated.
    Issues are found by labels. Some of those are derived from the scanned_element. If given,
    ctx_labels are also included in search-query (i.e. an issue must have all of the given ctx_labels
    to be considered a match). All of those labels + any given extra_labels are assigned to the
    issue, both in case it is created anew, or updated. Extra labels that are not ignored via the
    preserve_labels_regexes argument are dropped.
    '''

    open_issues = tuple(
        enumerate_issues(
            scanned_element=scanned_element,
            issue_type=issue_type,
            known_issues=known_issues,
            state='open',
            extra_labels=ctx_labels,
        )
    )
    if (issues_count := len(open_issues)) > 1:
        raise RuntimeError(
            f'more than one open issue found for {unique_name_for_element(scanned_element)=}'
        )
    elif issues_count == 0:
        if extra_labels:
            extra_labels = set(extra_labels) | set(ctx_labels)
        else:
            extra_labels = set(ctx_labels)

        return _create_issue(
            scanned_element=scanned_element,
            issue_type=issue_type,
            extra_labels=extra_labels,
            repository=repository,
            body=body,
            title=title,
            assignees=assignees,
            assignees_statuses=assignees_statuses,
            milestone=milestone,
            failed_milestones=failed_milestones,
            latest_processing_date=latest_processing_date,
        )
    elif issues_count == 1:

        open_issue = open_issues[0] # we checked there is exactly one
        open_issue: github3.issues.issue.ShortIssue

        def labels_to_preserve():
            nonlocal preserve_labels_regexes

            # always keep ctx_labels
            ctx_label_regex = f'{_label_prefix_ctx}.*'
            preserve_labels_regexes = set(preserve_labels_regexes) | set({ctx_label_regex})

            for label in open_issue.original_labels:
                for r in preserve_labels_regexes:
                    if re.fullmatch(pattern=r, string=label.name):
                        yield label.name
                        break

        if extra_labels:
            extra_labels = set(extra_labels) | set(labels_to_preserve())
        else:
            extra_labels = labels_to_preserve()
        try:
            return _update_issue(
                scanned_element=scanned_element,
                issue_type=issue_type,
                extra_labels=extra_labels,
                body=body,
                title=title,
                assignees=assignees,
                milestone=milestone,
                issue=open_issue,
            )
        except Exception as e:
            message = f'```\n{traceback.format_exc()}\n```'
            if isinstance(e, github3.exceptions.GitHubError) and e.errors: # noqa: E1101
                # also add much more helpful "errors", if available
                message += f'Additional error-details supplied by github: \n{e.errors}' # noqa: E1101
            _create_or_update_problems_comment(issue=open_issue, message=message)
            raise

    else:
        raise RuntimeError('this line should never be reached') # all cases should be handled before


@github.retry.retry_and_throttle
def close_issue_if_present(
    scanned_element: gcm.Target,
    issue_type: str,
    repository: github3.repos.Repository,
    known_issues: typing.Iterable[github3.issues.issue.ShortIssue],
    ctx_labels: typing.Iterable[str]=(),
):
    open_issues = tuple(
        enumerate_issues(
            scanned_element=scanned_element,
            issue_type=issue_type,
            known_issues=known_issues,
            state='open',
            extra_labels=ctx_labels,
        )
    )
    open_issues: tuple[github3.issues.ShortIssue]

    logger.info(f'{len(open_issues)=} found for closing {open_issues=}')

    if (issues_count := len(open_issues)) > 1:
        logger.warning(
            f'more than one open issue found for {unique_name_for_element(scanned_element)=}'
        )
    elif issues_count == 0:
        logger.info(f'no open issue found for {unique_name_for_element(scanned_element)=}')
        return # nothing to do

    open_issue = open_issues[0]
    logger.info(f'labels for issue for closing: {[l.name for l in open_issue.original_labels]}')

    for issue in open_issues:
        issue: github3.issues.ShortIssue
        issue.create_comment('closing ticket, because there are no longer unassessed findings')
        if not github.util.close_issue(issue):
            logger.warning(f'failed to close {issue.id=}, {repository.url=}')

    return issue
